# coding: utf-8

"""
Copyright 2016 SmartBear Software

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Ref: https://github.com/swagger-api/swagger-codegen
"""

#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems


class MlAssetsCreateExperimentInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, tags=None, name=None, description=None, framework=None, training_data_reference=None, author=None):
        """
        MlAssetsCreateExperimentInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'tags': 'list[TagRepository]',
            'name': 'str',
            'description': 'str',
            'framework': 'FrameworkOutputRepository',
            'training_data_reference': 'ArrayDataInputRepository',
            'author': 'AuthorRepository'
        }

        self.attribute_map = {
            'tags': 'tags',
            'name': 'name',
            'description': 'description',
            'framework': 'framework',
            'training_data_reference': 'training_data_reference',
            'author': 'author'
        }
        self._tags = tags
        self._name = name
        self._description = description
        self._framework = framework
        self._training_data_reference = training_data_reference
        self._author = author

    @property
    def tags(self):
        """
        Gets the tags of this ExperimentInputExperiments.


        :return: The tags of this ExperimentInputExperiments.
        :rtype: list[TagRepository]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of this ExperimentInputExperiments.


        :param tags: The tags of this ExperimentInputExperiments.
        :type: list[TagRepository]
        """

        self._tags = tags

    @property
    def name(self):
        """
        Gets the name of this MlAssetsCreateExperimentInput.


        :return: The name of this MlAssetsCreateExperimentInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this MlAssetsCreateExperimentInput.


        :param name: The name of this MlAssetsCreateExperimentInput.
        :type: str
        """
        self._name = name

    @property
    def description(self):
        """
        Gets the description of this MlAssetsCreateExperimentInput.


        :return: The description of this MlAssetsCreateExperimentInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this MlAssetsCreateExperimentInput.


        :param description: The description of this MlAssetsCreateExperimentInput.
        :type: str
        """
        self._description = description

    @property
    def framework(self):
        """
        Gets the framework of this MlAssetsCreateExperimentInput.


        :return: The framework of this MlAssetsCreateExperimentInput.
        :rtype: FrameworkOutputRepository
        """
        return self._framework

    @framework.setter
    def framework(self, framework):
        """
        Sets the framework of this MlAssetsCreateExperimentInput.


        :param framework: The framework of this MlAssetsCreateExperimentInput.
        :type: FrameworkOutputRepository
        """
        self._framework = framework

    @property
    def training_data_reference(self):
        """
        Gets the training_data of this MlAssetsCreateExperimentInput.


        :return: The training_data of this MlAssetsCreateExperimentInput.
        :rtype: ArrayDataInputRepository
        """
        return self._training_data_reference

    @training_data_reference.setter
    def training_data_reference(self, training_data_reference):
        """
        Sets the training_data of this MlAssetsCreateExperimentInput.


        :param training_data: The training_data of this MlAssetsCreateExperimentInput.
        :type: ArrayDataInputRepository
        """
        self._training_data_reference = training_data_reference

    @property
    def author(self):
        """
        Gets the author of this MlAssetsCreateExperimentInput.


        :return: The author of this MlAssetsCreateExperimentInput.
        :rtype: AuthorRepository
        """
        return self._author

    @author.setter
    def author(self, author):
        """
        Sets the author of this MlAssetsCreateExperimentInput.


        :param author: The author of this MlAssetsCreateExperimentInput.
        :type: AuthorRepository
        """
        self._author = author

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import logging
import time
import json
import numpy as np
import tensorflow as tf
import inspect
# if tf.__version__ != "2.1.0":
# raise ImportError("This function requires TensorFlow v2.1.0.")

from ibmfl.util import config
from ibmfl.util import fl_metrics
from ibmfl.model.fl_model import FLModel
from ibmfl.model.model_update import ModelUpdate
from ibmfl.exceptions import FLException, LocalTrainingException, ModelException

logger = logging.getLogger(__name__)


class TensorFlowFLModel(FLModel):
    """
    Wrapper class for importing tensorflow models.
    """

    def __init__(self, model_name, model_spec, tf_model=None, **kwargs):
        """
        Create a `TensorFlowFLModel` instance from a tensorflow model.\
        If `tf_model` is provided, it will use it; otherwise it will take\
        the model_spec to create the model.\
        Assumes the `tf_model` passed as argument is compiled.

        :param model_name: String specifying the type of model e.g., tf_CNN
        :type model_name: `str`
        :param model_spec: Specification of the `tf_model`
        :type model_spec: `dict`
        :param tf_model: Compiled TensorFlow model.
        :type tf_model: `tf.keras.Model`
        """

        super().__init__(model_name, model_spec, **kwargs)
        if tf_model is None:
            if model_spec is None or (not isinstance(model_spec, dict)):
                raise ValueError('Initializing model requires '
                                 'a model specification or '
                                 'compiled TensorFlow model. '
                                 'None was provided')
            # In this case we need to recreate the model from model_spec
            self.model = self.load_model_from_spec(model_spec)
        else:
            if not issubclass(type(tf_model), tf.keras.Model):
                raise ValueError('Compiled TensorFlow model needs to be '
                                 'provided of type `tensorflow.keras.models`.'
                                 ' Type provided: ' + str(type(tf_model)))

            if self.use_gpu_for_training and self.num_gpus >= 1:
                strategy = tf.distribute.MirroredStrategy()
                with strategy.scope():
                    self.model = tf_model
            else:
                self.model = tf_model
        self.model_type = 'TensorFlow-2.1.0'
        # Default values for local training
        self.batch_size = 128
        self.epochs = 1
        self.steps_per_epoch = None
        self.is_classification = True if not (model_spec and model_spec.get(
            'is_classification')) else model_spec.get('is_classification')

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data, a tuple\
        given in the form (x_train, y_train).
        :type train_data: `np.ndarray`
        :param fit_params: (optional) Dictionary with hyperparameters\
        that will be used to call fit function.\
        Hyperparameter parameters should match  expected values\
        e.g., `epochs`, which specifies the number of epochs to be run.\
        If no `epochs` or `batch_size` are provided, a default value\
        will be used (1 and 128, respectively).
        :type fit_params: `dict`
        :return: None
        """
        fit_args = self.get_fit_args(fit_params, **kwargs)

        logger.info('Training hps for this round => '
                    'batch_size: {}, epochs {}, steps_per_epoch {}'
                    .format(fit_args.get('batch_size'), 
                    fit_args.get('epochs'), fit_args.get('steps_per_epoch')))

        try:
            if type(train_data) is tuple and type(train_data[0]) is np.ndarray:
                # Extract x_train and y_train, by default,
                # label is stored in the last column
                x = train_data[0]
                y = train_data[1]
                self.model.fit(x, y, **fit_args)
            else:
                if isinstance(train_data, (tf.keras.utils.Sequence)) and \
                    hasattr(train_data, 'set_batch_size'):
                    train_data.set_batch_size(fit_args.get('batch_size'))

                fit_args.pop('batch_size', None)
                self.model.fit(train_data, **fit_args)

        except Exception as e:
            logger.exception(str(e))
            raise LocalTrainingException(
                'Error occurred while performing model.fit')

    def update_model(self, model_update):
        """
        Update TensorFlow model with provided model_update, where model_update \
        should be generated according to \
        `TensorFlowFLModel.get_model_update()`.

        :param model_update: `ModelUpdate` object that contains the weights \
        that will be used to update the model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if isinstance(model_update, ModelUpdate):
            w = model_update.get("weights")
            self.model.set_weights(w)
        else:
            raise LocalTrainingException('Provided model_update should be of '
                                         'type ModelUpdate. '
                                         'Instead they are: ' +
                                         str(type(model_update)))

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        w = self.model.get_weights()
        return ModelUpdate(weights=w)

    def predict(self, x, **kwargs):
        """
        Perform prediction for a batch of inputs. Note that for classification \
        problems, it returns the resulting probabilities.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param kwargs: Dictionary of tf-specific arguments.
        :type kwargs: `dict`

        :return: Array of predictions
        :rtype: `np.ndarray`
        """
        return self.model.predict(x, **kwargs)

    def evaluate(self, test_dataset, **kwargs):
        """
        Evaluates the model given testing data.

        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, y_test) or a datagenerator of type `keras.utils.Sequence`, \
        `keras.preprocessing.image.ImageDataGenerator`
        :type test_dataset: `np.ndarray`
        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        :return: metrics
        :rtype: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test, **kwargs)

        else:
            return self.evaluate_generator_model(
                test_dataset, **kwargs)

    def evaluate_model(self, x, y, batch_size=128, **kwargs):
        """
        Evaluates the model given x and y.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding labels to x
        :type y: `np.ndarray`
        :param batch_size: Size of batches.
        :type batch_size: `int`
        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        :return: metrics
        :rtype: `dict`
        """

        metrics = self.model.evaluate(
            x, y, batch_size=batch_size, **kwargs)
        names = self.model.metrics_names
        dict_metrics = {}
        additional_metrics = {}
        if type(metrics) == list:
            for metric, name in zip(metrics, names):
                # metric = metric.item()
                if name == 'accuracy':
                    dict_metrics['acc'] = round(metric, 2)
                dict_metrics[name] = metric
        else:
            dict_metrics[names[0]] = metrics

        y_pred = self.predict(x, batch_size=batch_size)
        if self.is_classification:
            additional_metrics = fl_metrics.get_eval_metrics_for_classificaton(
                y, y_pred)
        else:
            additional_metrics = fl_metrics.get_eval_metrics_for_regression(
                y, y_pred)

        logger.info(additional_metrics)
        dict_metrics = {**dict_metrics, **additional_metrics}
        logger.info(dict_metrics)

        return dict_metrics

    def evaluate_generator_model(self, test_generator, **kwargs):
        """
        Evaluates the model based on the provided data generator.

        :param test_generator: Testing datagenerator of type \
        `keras.utils.Sequence`, or \
        `keras.preprocessing.image.ImageDataGenerator`.
        :type test_generator: `ImageDataGenerator` or `keras.utils.Sequence`
        :return: metrics
        :rtype: `dict`
        """

        steps = self.steps_per_epoch
        if steps in kwargs:
            steps = kwargs.get('steps')

        metrics = self.model.evaluate_generator(
            test_generator, steps=steps)
        names = self.model.metrics_names
        dict_metrics = {}
        additional_metrics = {}

        if type(metrics) == list:
            for metric, name in zip(metrics, names):
                # metric = metric.item()
                if name == 'accuracy':
                    dict_metrics['acc'] = round(metric, 2)
                dict_metrics[name] = metric
        else:
            dict_metrics[names[0]] = metrics

        return dict_metrics

    @staticmethod
    def load_model(file_name, custom_objects={}):
        """
        Loads a model from disk given the specified file_name

        :param file_name: Name of the file that contains the model to be loaded.
        :type file_name: `str`
        :return: TensorFlow model loaded to memory
        :rtype: `tf.keras.models.Model`
        """
        try:
            model = tf.keras.models.load_model(
                file_name, custom_objects=custom_objects)
        except Exception as ex:
            logger.exception(str(ex))
            logger.error(
                'Loading model via tf.keras.models.load_model failed!')
        return model

    def save_model(self, filename=None):
        """
        Save a model to file in the format specific to the backend framework.

        :param filename: Name of the file where to store the model.
        :type filename: `str`
        :return: filename
        :rtype `string`
        """
        if filename is None:
            file = self.model_name if self.model_name else self.model_type
            filename = '{}'.format(file)

        full_path = super().get_model_absolute_path(filename)
        self.model.save(full_path)
        logger.info('Model saved in path: %s.', full_path)
        return filename

    @staticmethod
    def model_from_json_via_tf_keras(json_file_name, custom_objects={}):
        """
        Loads a model architecture from disk via tf.keras \
        given the specified json file name.

        :param json_file_name: Name of the file that contains \
        the model architecture to be loaded.
        :type json_file_name: `str`
        :param custom_objects: Dictionary of custom objects required for loading arch
        :type custom_objects: `dict`
        :return: tf.keras model with only model architecture loaded to memory
        :rtype: `tf.keras.models.Model`
        """
        model = None
        json_file = open(json_file_name, 'r')
        f = json_file.read()
        json_file.close()
        try:
            model = tf.keras.models.model_from_json(
                f, custom_objects=custom_objects)
        except Exception as ex:
            logger.error(
                'Loading model via tf.keras.models.model_from_json failed! ')

        return model

    def load_model_from_spec(self, model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict` \
        that contains the following items: \
            'model_definition': the path where the tf model is stored, \
                usually in a `SavedModel` format.
        :return: model
        :rtype: `keras.models.Model`
        """
        custom_objects = {}
        if 'custom_objects' in model_spec:

            custom_objects_config = model_spec['custom_objects']
            for custom_object in custom_objects_config:
                key = custom_object['key']
                value = custom_object['value']
                path = custom_object['path']
                custom_objects[key] = config.get_attr_from_path(
                    path, value)

        if 'model_definition' in model_spec:
            try:
                model_file = model_spec['model_definition']
                model_absolute_path = config.get_absolute_path(model_file)

                if self.use_gpu_for_training:
                    strategy = tf.distribute.MirroredStrategy()
                    with strategy.scope():
                        model = TensorFlowFLModel.load_model(
                            model_absolute_path, custom_objects=custom_objects)
                else:
                    model = TensorFlowFLModel.load_model(
                        model_absolute_path, custom_objects=custom_objects)

            except Exception as ex:
                logger.exception(str(ex))
                raise FLException('Failed to load TensorFlow model!')
        else:

            if self.use_gpu_for_training:
                strategy = tf.distribute.MirroredStrategy()
                model = self.load_model_from_architecture(
                    model_spec, custom_objects)

            else:
                model = self.load_model_from_architecture(
                    model_spec, custom_objects)

        return model

    def load_model_from_architecture(self, model_spec, custom_objects):
        """
        Loads model from provided model_spec, where model_spec is a `dict` \
        that contains the following items: \
            'model_architecture': the path where the tf model is stored, \
                usually in a `SavedModel` format. \
            'model_weights': the path to where the tf model weights are saved \
            'compile_model_options': attributes used to compile model.
        :param model_spec: Disctionary of spec provided by the user
        :type model_spec: `dict`
        :param custom_objects: Dictionary of custom objects required for loading arch
        :type custom_objects: `dict`
        :return: model
        :rtype: `keras.models.Model`
        """

        try:
            model = TensorFlowFLModel.model_from_json_via_tf_keras(
                model_spec['model_architecture'], custom_objects=custom_objects)

            if model is None:
                logger.error('An acceptable compiled model should be of type '
                             'tensorflow.keras.models!')
        except Exception as ex:
            logger.error(str(ex))
            raise FLException('Unable to load the provided uncompiled model!')

            # Load weights from provided path
        if 'model_weights' in model_spec:
            model.load_weights(model_spec['model_weights'])

        if 'compile_model_options' in model_spec:
            # Load compile options:
            try:
                compiled_options = model_spec['compile_model_options']
                optimizer = self.get_custom_attribute(
                    compiled_options.get('optimizer'))
                loss = self.get_custom_attribute(compiled_options.get('loss'))
                metrics = self.get_custom_attribute(
                    compiled_options.get('metrics'))
                metrics = [metrics] if not isinstance(
                    metrics, list) else metrics
                model.compile(optimizer=optimizer,
                              loss=loss,
                              metrics=metrics)
            except Exception as ex:
                logger.exception(str(ex))
                logger.exception(
                    'Failed to compiled the TensorFlow.keras model.')
        else:
            raise ModelException('Failed to compile keras model, '
                                 'no compile options provided.')

        return model

    def get_gradient(self, train_data):
        """
        Compute the gradient with the provided dataset at the current local \
        model's weights.

        :param train_data: Training data, a tuple \
        given in the form (x_train, y_train).
        :type train_data: `np.ndarray`
        :return: gradients
        :rtype: `list` of `tf.Tensor`
        """
        try:
            x, y = train_data[0], train_data[1]
        except Exception as ex:
            logger.exception(str(ex))
            raise FLException('Provided dataset has incorrect format. '
                              'It should be a tuple in the form of '
                              '(x_train, y_train).')
        with tf.GradientTape() as tape:
            predictions = self.model(x, training=True)
            loss = self.model.loss(y, predictions)

        gradients = tape.gradient(loss, self.model.trainable_variables)
        return gradients

    def expand_model_by_layer_name(self, new_dimension, layer_name="dense"):
        """
        Expand the current Keras model with provided dimension of
        the hidden layers or model weights.
        This method by default expands the dense layer of
        the current neural network.
        It can be extends to expand other layers specified by `layer_name`,
        for example, it can be use to increase the number of CNN filters or
        increase the hidden layer size inside LSTM.

        :param new_dimension: New number of dimensions for \
        the fully connected layers
        :type new_dimension: `list`
        :param layer_name: layer's name to be expanded
        :type layer_name: `str`
        :return: None
        """
        if new_dimension is None:
            raise FLException('No information is provided for '
                              'the new expanded model. '
                              'Please provide the new dimension of '
                              'the resulting expanded model.')
        try:
            model_config = json.loads(self.model.to_json())
        except NotImplementedError:
            raise ModelException(
                "Please construct the model config for models in "
                "`SavedModel` format. "
                "Details about how to construct the model config can be found"
                " in TensorFlowFLModel tutorials.")
        except Exception as ex:
            logger.exception(str(ex))
            raise FLException("Error occurred during extracting "
                              "the model architecture.")
        i = 0

        for layer in model_config['config']['layers']:
            # find the specified layers
            if 'class_name' in layer and \
                    layer['class_name'].strip().lower() == layer_name:
                layer['config']['units'] = new_dimension[i]
                i += 1

        custom_obj = {
            self.model.__class__.__name__: self.model.__class__
        }

        try:
            new_model = tf.keras.models.model_from_json(
                json.dumps(model_config), custom_objects=custom_obj)
        except Exception as ex:
            logger.exception(str(ex))
            raise FLException("Error occurred during loading model from "
                              "the new config.")

        metrics = self.model.metrics_names
        if 'loss' in metrics:
            metrics.remove('loss')
        if not self.use_gpu_for_training or self.num_gpus == 1:
            new_model.compile(optimizer=self.model.optimizer,
                              loss=self.model.loss,
                              metrics=metrics)
        else:
            strategy = tf.distribute.MirroredStrategy()
            with strategy.scope():
                new_model.compile(optimizer=self.model.optimizer,
                                  loss=self.model.loss,
                                  metrics=metrics)

        self.model = new_model

    def is_fitted(self):
        """
        Return a boolean value indicating if the model is fitted or not. \
        In particular, check if the tensorflow model has weights. \
        If it has, return True; otherwise return false. 

        :return: res
        :rtype: `bool`
        """
        try:
            self.model.get_weights()
        except Exception:
            return False
        return True

    def get_loss(self, dataset):
        """
        Return the resulting loss computed based on the provided dataset.

        :param dataset: Provided dataset, a tuple given in the form \
        (x_test, y_test) or a datagenerator of type `keras.utils.Sequence`, \
        `keras.preprocessing.image.ImageDataGenerator`.
        :type dataset: `np.ndarray`
        :return: The resulting loss.
        :rtype: `float`
        """
        if 'loss' not in self.model.metrics_names:
            self.model.metrics_names.append('loss')
        res = self.evaluate(dataset)

        if 'loss' in res:
            loss = round(res['loss'], 2)
            return loss
        else:
            raise FLException(
                "Loss is not listed in the model's metrics_names.")

    def get_custom_attribute(self, attr):
        """
        Load compiled options which are provided as config.
        :param attr: Attribute config provided in config
        :type attr: dict or key
        :return: Attribute loaded and returned back for compilation
        :rtype: `str` or python attr

        """
        if attr is None:
            raise ModelException("Invalid Model config exception")

        if isinstance(attr, dict):
            try:

                value = attr.get('value')
                path = attr.get('path')
                args = attr.get('args') if 'args' in attr else {}
                attribute = config.get_attr_from_path(
                    path, value)

            except Exception as ex:
                logger.error(
                    "Error occurred while loading the custom attribute!")
                logger.error("Custom attribute : " + attr)
                logger.error()

            logger.debug(type(attribute))

            if inspect.isclass(attribute):
                return attribute(**args)
            else:
                return attribute

        else:
            return attr

    def get_fit_args(self, global_params, **kwargs):

        fit_args = {}
        local_params = kwargs.get('local_params', {}) or {} if global_params else {}

        hyperparams = global_params.get('hyperparams', {}) or {} if global_params else {}
        local_hp = hyperparams.get('local', {}) or {}
        training_hp = local_hp.get('training', {}) or {}

        fit_args['batch_size'] = training_hp.get('batch_size', self.batch_size)
        if 'batch_size' in local_params:
            fit_args['batch_size'] = local_params.get('batch_size')

        fit_args['epochs'] = training_hp.get('epochs', self.epochs)
        if 'epochs' in local_params:
            fit_args['epochs'] = local_params.get('epochs')

        fit_args['steps_per_epoch'] = training_hp.get('steps_per_epoch', self.steps_per_epoch)
        if 'steps_per_epoch' in local_params:
            fit_args['steps_per_epoch'] = local_params.get('steps_per_epoch')

        return fit_args



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
import sys

from ibm_watson_machine_learning.utils.autoai.utils import try_import_tqdm

try_import_tqdm()

from tqdm import tqdm as TQDM

__all__ = [
    "ProgressBar"
]


class ProgressBar(TQDM):
    """
    Progress Bar class for handling progress bar display. It is based on 'tqdm' class, could be extended.

    Parameters
    ----------
    desc: str, optional
        Description string to be added as a prefix to progress bar.

    total: int, optional
        Total length of the progress bar.
    """

    def __init__(self, ncols: Union[str, int], position: int = 0, desc: str = None, total: int = 100,
                 leave: bool = True, bar_format: str = '{desc}: {percentage:3.0f}%|{bar}|') -> None:
        # note: to see possible progress bar formats please look at super 'bar_format' description
        super().__init__(desc=desc, total=total, leave=leave, position=position, ncols=ncols, file=sys.stdout,
                         bar_format=bar_format)
        self.total = total
        self.previous_message = None
        self.counter = 0
        self.progress = 0

    def increment_counter(self, progress: int = 5) -> None:
        """
        Increments internal counter and waits for specified time.

        Parameters
        ----------
        progress: int, optional
            How many steps at a time progress bar will increment.
        """
        self.progress = progress
        self.counter += progress

    def reset_counter(self) -> None:
        """
        Restart internal counter
        """
        self.counter = 0

    def update(self):
        """
        Updates the counter with specific progress.
        """
        super().update(n=self.progress)

    def last_update(self):
        """Fill up the progress bar till the end, this was the last run."""
        super().update(n=self.total - self.counter)



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ibm_watson_machine_learning.libs.repo.mlrepository import  WmlExperimentArtifact
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps

from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker

lib_checker = LibraryChecker()

logger = logging.getLogger('WmlExperimentCollectionAdapter')


class WmlExperimentCollectionAdapter(object):
    """
    Adapter creating experiment artifact using output from service.
    """

    def __init__(self, experiment_output, client):

        self.experiment_output = experiment_output
        self.client = client
        self.experiment_entity = experiment_output.entity
        self.experiment_metadata = experiment_output.metadata

    def artifact(self):
        experiement_artifact_builder = type(
                "ExperimentArtifact",
                (WmlExperimentArtifact, object),
                {}
            )

        prop_map = {
            MetaNames.CREATION_TIME: self.experiment_metadata.created_at,
            MetaNames.LAST_UPDATED: self.experiment_metadata.modified_at,
            MetaNames.EXPERIMENT_URL : self.experiment_metadata.url
        }

        if self.experiment_entity.settings is not None:
            prop_map[MetaNames.EXPERIMENTS.SETTINGS] = self.experiment_entity.settings


        if self.experiment_entity.tags is not None:
            prop_map[MetaNames.EXPERIMENTS.TAGS] = self.experiment_entity.tags


        if self.experiment_entity.training_data_reference is not None:
            prop_map[MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE] = self.experiment_entity.training_data_reference

        if self.experiment_entity.training_references is not None:
            prop_map[MetaNames.EXPERIMENTS.TRAINING_REFERENCES] = self.experiment_entity.training_references

        if self.experiment_entity.training_results_reference is not None:
            prop_map[MetaNames.EXPERIMENTS.TRAINING_RESULTS_REFERENCE] = self.experiment_entity.training_results_reference


        name = None
        experiment_url = self.experiment_metadata.url
        experiment_id = experiment_url.split("/experiments/")[1].split("/")[0]


        experiment_artifact = experiement_artifact_builder(
            experiment_id,
            name,
            MetaProps(prop_map)
        )

        experiment_artifact.client = self.client


        return experiment_artifact



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.utils import version

#from ibm_watson_machine_learning.learning_system import LearningSystem
from ibm_watson_machine_learning.experiments import Experiments
from ibm_watson_machine_learning.repository import Repository
from ibm_watson_machine_learning.model_definition import ModelDefinition
from ibm_watson_machine_learning.models import Models
from ibm_watson_machine_learning.pipelines import Pipelines
from ibm_watson_machine_learning.deployments import Deployments
from ibm_watson_machine_learning.factsheets import Factsheets
from ibm_watson_machine_learning.training import Training
from ibm_watson_machine_learning.functions import Functions
from ibm_watson_machine_learning.platform_spaces import PlatformSpaces
from ibm_watson_machine_learning.assets import Assets
from ibm_watson_machine_learning.connections import Connections
from ibm_watson_machine_learning.Set import Set
from ibm_watson_machine_learning.sw_spec import SwSpec
from ibm_watson_machine_learning.hw_spec import HwSpec
from ibm_watson_machine_learning.pkg_extn import PkgExtn
from ibm_watson_machine_learning.shiny import Shiny
from ibm_watson_machine_learning.script import Script
from ibm_watson_machine_learning.wml_client_error import NoWMLCredentialsProvided
from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.instance_new_plan import ServiceInstanceNewPlan

from ibm_watson_machine_learning.remote_training_system import RemoteTrainingSystem
from ibm_watson_machine_learning.volumes import Volume

import copy
import os
import requests
import sys


'''
.. module:: APIClient
   :platform: Unix, Windows
   :synopsis: Watson Machine Learning API Client.

.. moduleauthor:: IBM
'''


class APIClient:

    def __init__(self, wml_credentials, project_id=None):
        self._logger = logging.getLogger(__name__)
        self.wml_credentials = copy.deepcopy(wml_credentials)
        self.CAMS = None
        self.WSD = None
        self.WSD_20 = None
        self.ICP_30 = None
        self.ICP = None
        self.default_space_id = None
        self.default_project_id = None
        self.project_type = None
        self.CLOUD_PLATFORM_SPACES = False
        self.PLATFORM_URL = None
        self.CAMS_URL = None
        self.CREATED_IN_V1_PLAN = False
        self.version_param = '2021-06-24'
        self.ICP_PLATFORM_SPACES = False # This will be applicable for 3.5 and later and specific to convergence functionalities
        self.ICP_35 = False # Use it for any 3.5 specific functionalities
        self.ICP_40 = False # Use it for any 4.0 specific functionalities
        self.CLOUD_BETA_FLOW = False

        self.PLATFORM_URLS_MAP = {
            'https://wml-dev.ml.test.cloud.ibm.com': 'https://api.dataplatform.dev.cloud.ibm.com',
            'https://wml-fvt.ml.test.cloud.ibm.com': 'https://api.dataplatform.dev.cloud.ibm.com',
            'https://us-south.ml.test.cloud.ibm.com': 'https://api.dataplatform.dev.cloud.ibm.com',
            'https://yp-qa.ml.cloud.ibm.com': 'https://api.dataplatform.test.cloud.ibm.com',
            'https://private.yp-qa.ml.cloud.ibm.com': 'https://api.dataplatform.test.cloud.ibm.com',
            'https://yp-cr.ml.cloud.ibm.com': 'https://api.dataplatform.test.cloud.ibm.com',
            'https://private.yp-cr.ml.cloud.ibm.com': 'https://api.dataplatform.test.cloud.ibm.com',
            'https://jp-tok.ml.cloud.ibm.com': 'https://api.jp-tok.dataplatform.cloud.ibm.com',
            'https://eu-gb.ml.cloud.ibm.com': 'https://api.eu-gb.dataplatform.cloud.ibm.com',
            'https://eu-de.ml.cloud.ibm.com': 'https://api.eu-de.dataplatform.cloud.ibm.com',
            'https://us-south.ml.cloud.ibm.com': 'https://api.dataplatform.cloud.ibm.com'
        }

        import requests
        requests.packages.urllib3.disable_warnings()

        if "token" in wml_credentials:
            self.proceed = True
        else:
            self.proceed = False

        self.project_id = project_id
        self.wml_token = None
        self.__wml_local_supported_version_list = ['1.0', '1.1',
                                                   '2.0', '2.0.1', '2.0.2', '2.5.0',
                                                   '3.0.0', '3.0.1', '3.5', '4.0']
        self.__wsd_supported_version_list = ['1.0', '1.1', '2.0']
        self.__predefined_instance_type_list = ['icp', 'openshift', 'wml_local', 'wsd_local']
        os.environ['WSD_PLATFORM'] = 'False'
        if wml_credentials is None:
            raise NoWMLCredentialsProvided()
        if 'url' not in self.wml_credentials:
            raise WMLClientError(Messages.get_message(message_id="url_not_provided"))
        if not self.wml_credentials['url'].startswith('https://'):
            raise WMLClientError(Messages.get_message(message_id="invalid_url"))
        if self.wml_credentials['url'][-1] == "/":
            self.wml_credentials['url'] = self.wml_credentials['url'].rstrip('/')

        if 'instance_id' not in wml_credentials.keys():
            self.CLOUD_PLATFORM_SPACES = True
            self.wml_credentials[u'instance_id'] = 'invalid'  # This is applicable only via space instance_id
            self.ICP = False
            self.CAMS = False

            if wml_credentials[u'url'] in self.PLATFORM_URLS_MAP.keys():
                self.PLATFORM_URL = self.PLATFORM_URLS_MAP[wml_credentials[u'url']]
                self.CAMS_URL = self.PLATFORM_URLS_MAP[wml_credentials[u'url']]
            else:
                raise WMLClientError(Messages.get_message(message_id="invalid_cloud_scenario_url"))

            if not self._is_IAM():
                raise WMLClientError(Messages.get_message(message_id="apikey_not_provided"))
        else:
            if 'icp' == wml_credentials[u'instance_id'].lower() or 'openshift' == wml_credentials[
                u'instance_id'].lower() or 'wml_local' == wml_credentials[u'instance_id'].lower():
                self.ICP = True
                os.environ["DEPLOYMENT_PLATFORM"] = "private"
                ##Condition for CAMS related changes to take effect (Might change)
                if 'version' not in wml_credentials:
                    raise WMLClientError(Messages.get_message(self.__wml_local_supported_version_list, message_id="version_not_provided"))

                if 'version' in wml_credentials.keys() and (
                        '2.0.1' == wml_credentials[u'version'].lower() or '2.5.0' == wml_credentials[
                    u'version'].lower() or
                        '3.0.0' == wml_credentials[u'version'].lower() or '3.0.1' == wml_credentials[
                            u'version'].lower() or '3.5' == wml_credentials[u'version'].lower() or
                        '4.0' == wml_credentials[u'version'].lower() or
                        '1.1' == wml_credentials[u'version'].lower() or '2.0' == wml_credentials[u'version'].lower()):
                    self.CAMS = True
                    os.environ["DEPLOYMENT_PRIVATE"] = "icp4d"
                    if 'wml_local' == wml_credentials[u'instance_id'].lower() and \
                            ('1.1' == wml_credentials[u'version'].lower() or '2.0' == wml_credentials[
                                u'version'].lower()):
                        url_port = wml_credentials[u'url'].split(':')[-1]
                        if not url_port.isdigit():
                            raise WMLClientError(Messages.get_message(message_id="no_port_number_for_wml_local"))

                    if '3.0.0' == wml_credentials[u'version'].lower() or \
                            '3.0.1' == wml_credentials[u'version'].lower() or \
                            '2.0' == wml_credentials[u'version'].lower():
                        self.ICP_30 = True

                    # For Cloud convergence related functionalities brought into CP4D 3.5
                    if '3.5' == wml_credentials[u'version'].lower():
                        self.ICP_PLATFORM_SPACES = True
                        self.ICP_35 = True

                    if '4.0' == wml_credentials[u'version'].lower():
                        self.ICP_PLATFORM_SPACES = True
                        self.ICP_40 = True
                        if 'bedrock_url' not in wml_credentials:
                            self.wml_credentials['bedrock_url'] = '.'.join(['https://cp-console'] + wml_credentials['url'].split('.')[1:])
                            self._is_bedrock_url_autogenerated = True

                else:
                    if 'version' in wml_credentials.keys() and \
                            wml_credentials[u'version'].lower() not in self.__wml_local_supported_version_list:
                        raise WMLClientError(Messages.get_message(
                            ', '.join(self.__wml_local_supported_version_list),
                            message_id="invalid_version"))

                    self.CAMS = False
            else:
                if ('wsd_local' == wml_credentials[u'instance_id'].lower()) and \
                        ('1.1' == wml_credentials[u'version'].lower() or '2.0' == wml_credentials[u'version'].lower()):
                    self.WSD = True
                    os.environ['WSD_PLATFORM'] = 'True'
                    if '2.0' == wml_credentials[u'version'].lower():
                        self.WSD_20 = True
                else:
                    if ('wsd_local' == wml_credentials[u'instance_id'].lower()) and \
                            'version' in wml_credentials.keys() and \
                            wml_credentials[u'version'].lower() not in self.__wsd_supported_version_list:
                        raise WMLClientError(Messages.get_message(
                            ', '.join(self.__wsd_supported_version_list),
                            message_id="invalid_version"))

                    if 'instance_id' in wml_credentials.keys():
                        if wml_credentials['url'] in self.PLATFORM_URLS_MAP:
                            raise WMLClientError(Messages.get_message(message_id="instance_id_in_cloud_scenario"))
                        else:
                            raise WMLClientError(Messages.get_message(message_id="invalid_instance_id"))

                    self.ICP = False
                    self.CAMS = False

                    self.service_instance = ServiceInstanceNewPlan(self)

                    headers = self._get_headers()

                    del headers[u'X-WML-User-Client']
                    if 'ML-Instance-ID' in headers:
                        del headers[u'ML-Instance-ID']
                    del headers[u'x-wml-internal-switch-to-new-v4']

                    response_get_instance = requests.get(
                        u'{}/ml/v4/instances/{}'.format(self.wml_credentials['url'],
                                                  self.wml_credentials['instance_id']),
                        params={'version': self.version_param},
                        # params=self._client._params(),
                        headers=headers
                    )

                    if response_get_instance.status_code == 200:
                        if 'plan' in response_get_instance.json()[u'entity']:
                            tags = response_get_instance.json()[u'metadata']['tags']
                            if 'created_in_v1_plan' in tags:
                                self.CREATED_IN_V1_PLAN = True
                            if response_get_instance.json()[u'entity'][u'plan'][u'version'] == 2 and\
                                    not self.CREATED_IN_V1_PLAN and\
                                    response_get_instance.json()[u'entity'][u'plan']['name'] != 'lite':
                                self.CLOUD_PLATFORM_SPACES = True

                                # If v1 lite converted to v2 plan, tags : ["created_in_v1_plan"]

                                if not self._is_IAM():
                                    raise WMLClientError(Messages.get_message(message_id="apikey_not_provided"))

                                if wml_credentials[u'url'] in self.PLATFORM_URLS_MAP.keys():
                                    self.PLATFORM_URL = self.PLATFORM_URLS_MAP[wml_credentials[u'url']]
                                    self.CAMS_URL = self.PLATFORM_URLS_MAP[wml_credentials[u'url']]
                                else:
                                    raise WMLClientError(Messages.get_message(message_id="url_not_provided"))

                                if self.CLOUD_PLATFORM_SPACES:
                                    print(Messages.get_message(message_id="instance_id_is_not_required_for_v2_plans"))

                            if not self.CLOUD_PLATFORM_SPACES:
                                self.CLOUD_BETA_FLOW = True
                                print(Messages.get_message(message_id="creating_assets_using_v4_beta_flow_is_deprecated"))
                    elif response_get_instance.status_code == 404:
                        raise WMLClientError(Messages.get_message(message_id="instance_id_not_found"))
                    elif response_get_instance.status_code == 401:
                        raise WMLClientError(Messages.get_message(message_id="not_authorized_to_access_the_instance_id"))
                    else:
                        # if someone is using beta flow and using instance api key
                        response_get_instance_v1 = requests.get(
                            u'{}/v3/wml_instances/{}'.format(self.wml_credentials['url'],
                                                             self.wml_credentials['instance_id']),
                            headers=self._get_headers()
                        )

                        if response_get_instance_v1.status_code != 200:
                            raise WMLClientError(Messages.get_message(
                                response_get_instance.text,
                                message_id="failed_to_get_instance_details"))

                        self.CLOUD_BETA_FLOW = True
                        print(Messages.get_message(message_id="creating_assets_using_v4_beta_flow_is_deprecated"))
            if "token" in wml_credentials:
                self.proceed = True
            else:
                self.proceed = False

        if 'instance_id' in wml_credentials.keys() and \
                (wml_credentials['instance_id'].lower() not in self.__predefined_instance_type_list) and \
                'version' in wml_credentials.keys():
            raise WMLClientError(Messages.get_message(message_id="provided_credentials_are_invalid"))

        self.project_id = project_id
        self.wml_token = None

        if not self.CLOUD_PLATFORM_SPACES and not self.ICP_PLATFORM_SPACES and not self.CLOUD_BETA_FLOW:
            raise WMLClientError(Messages.get_message(message_id="client_is_not_supported_in_this_release"))

        # if not self.ICP and not self.ICP_30 and not self.WSD:
        if not self.WSD and not self.CLOUD_PLATFORM_SPACES and not self.ICP_PLATFORM_SPACES:
            from ibm_watson_machine_learning.instance import ServiceInstance
            self.service_instance = ServiceInstance(self)
            self.service_instance.details = self.service_instance.get_details()

        if self.CLOUD_PLATFORM_SPACES or self.ICP_PLATFORM_SPACES:
            # For cloud, service_instance.details will be set during space creation( if instance is associated ) or
            # while patching a space with an instance
            import sys
            if (3 == sys.version_info.major) and (6 == sys.version_info.minor):
                if self.CLOUD_PLATFORM_SPACES:
                    print(Messages.get_message(message_id="python_3_6_framework_is_deprecated"))
                elif self.ICP_PLATFORM_SPACES:
                    print(Messages.get_message(message_id="python_3_6_framework_is_deprecated_cp4d"))

            self.service_instance = ServiceInstanceNewPlan(self)
            self.volumes = Volume(self)

            if self.ICP_PLATFORM_SPACES:
                self.service_instance.details = self.service_instance.get_details()

            self.set = Set(self)

            self.spaces = PlatformSpaces(self)

            from ibm_watson_machine_learning.export_assets import Export
            from ibm_watson_machine_learning.import_assets import Import
            self.export_assets = Export(self)
            self.import_assets = Import(self)

            if self.ICP_PLATFORM_SPACES:
                self.shiny = Shiny(self)

            self.script = Script(self)
            self.model_definitions = ModelDefinition(self)

            self.package_extensions = PkgExtn(self)
            self.software_specifications = SwSpec(self)

            self.hardware_specifications = HwSpec(self)

            self.connections = Connections(self)
            self.training = Training(self)

            self.data_assets = Assets(self)

            self.deployments = Deployments(self)

            if self.CLOUD_PLATFORM_SPACES:
                from ibm_watson_machine_learning.migration_v4ga_cloud import Migrationv4GACloud
                self.v4ga_cloud_migration = Migrationv4GACloud(self)
                self.factsheets = Factsheets(self)

            self.remote_training_systems = RemoteTrainingSystem(self)

        ##Initialize Assets and Model_Definitions only for CAMS
        if (self.CAMS or self.WSD) and not self.ICP_PLATFORM_SPACES:
            self.set = Set(self)
            self.data_assets = Assets(self)
            self.model_definitions = ModelDefinition(self)
            if self.ICP_30:
                self.connections = Connections(self)
                self.software_specifications = SwSpec(self)
                self.hardware_specifications = HwSpec(self)
                self.package_extensions = PkgExtn(self)
                self.script = Script(self)
                if not '2.0' == wml_credentials[u'version'].lower():
                    self.shiny = Shiny(self)
            if self.WSD_20:
                self.software_specifications = SwSpec(self)

        #    self.learning_system = LearningSystem(self)
        self.repository = Repository(self)
        self._models = Models(self)
        self.pipelines = Pipelines(self)
        self.experiments = Experiments(self)
        self._functions = Functions(self)

        if not self.WSD and not self.CLOUD_PLATFORM_SPACES and not self.ICP_PLATFORM_SPACES:
            from ibm_watson_machine_learning.runtimes import Runtimes
            from ibm_watson_machine_learning.spaces import Spaces
            self.runtimes = Runtimes(self)
            self.deployments = Deployments(self)
            self.training = Training(self)
            self.spaces = Spaces(self)
            self.connections = Connections(self)

        self._logger.info(Messages.get_message(message_id="client_successfully_initialized"))
        self.version = version()

    def _check_if_either_is_set(self):
        if self.CAMS or self.CLOUD_PLATFORM_SPACES:
            if self.default_space_id is None and self.default_project_id is None:
                raise WMLClientError(Messages.get_message(message_id="it_is_mandatory_to_set_the_space_project_id"))

    def _check_if_space_is_set(self):
        if self.CAMS or self.CLOUD_PLATFORM_SPACES:
            if self.default_space_id is None:
                raise WMLClientError(Messages.get_message(message_id="it_is_mandatory_to_set_the_space_id"))

    def _params(self, skip_space_project_chk=False, skip_for_create=False):
        params = {}
        if self.CAMS or self.CLOUD_PLATFORM_SPACES:
            if self.CLOUD_PLATFORM_SPACES or self.ICP_PLATFORM_SPACES:
                params.update({'version': self.version_param})
            if not skip_for_create:
                if self.default_space_id is not None:
                    params.update({'space_id': self.default_space_id})
                elif self.default_project_id is not None:
                    params.update({'project_id': self.default_project_id})
                else:
                    # For system software/hardware specs
                    if skip_space_project_chk is False:
                        raise WMLClientError(Messages.get_message(message_id="it_is_mandatory_to_set_the_space_project_id"))

        if self.WSD:
            if self.default_project_id is not None:
                params.update({'project_id': self.default_project_id})
            else:
                raise WMLClientError(Messages.get_message(message_id="it_is_mandatory_to_set_the_project_id"))

        if self.project_type == 'local_git_storage':
            params.update({'userfs': 'true'})

        if self.project_type != 'local_git_storage' and 'userfs' in params:
            del params['userfs']

        return params

    def _get_headers(self, content_type='application/json', no_content_type=False, wsdconnection_api_req=False, zen=False):
        if self.WSD:
                headers = {'X-WML-User-Client': 'PythonClient'}
                if self.project_id is not None:
                    headers.update({'X-Watson-Project-ID': self.project_id})
                if not no_content_type:
                    headers.update({'Content-Type': content_type})
                if wsdconnection_api_req is True:
                    token = "desktop user token"
                else:
                    token = "desktop-token"
                headers.update({'Authorization':  "Bearer " + token})

        elif zen:
            headers= {'Content-Type': content_type}
            headers.update({'Authorization': "Bearer " + self.service_instance._create_token()})
        else:
            if self.proceed is True:
                token = "Bearer "+ self.wml_credentials["token"]
            else:
                token = "Bearer " + self.service_instance._get_token()
            headers = {
                'Authorization': token,
                'X-WML-User-Client': 'PythonClient'
            }
            # Cloud Convergence
            if self._is_IAM() or (self.service_instance._is_iam() is None and not self.CLOUD_PLATFORM_SPACES and not self.ICP_PLATFORM_SPACES):
                headers['ML-Instance-ID'] = self.wml_credentials['instance_id']

            headers.update({'x-wml-internal-switch-to-new-v4': "true"})
            if not self.ICP:
                #headers.update({'x-wml-internal-switch-to-new-v4': "true"})
                if self.project_id is not None:
                    headers.update({'X-Watson-Project-ID': self.project_id})

            if not no_content_type:
                headers.update({'Content-Type': content_type})

        return headers

    def _get_icptoken(self):
        return self.service_instance._create_token()

    def _is_default_space_set(self):
        if self.default_space_id is not None:
            return True
        return False

    def _is_IAM(self):
        if('apikey' in self.wml_credentials.keys()):
            if (self.wml_credentials['apikey'] != ''):
                return True
            else:
                raise WMLClientError(Messages.get_message(message_id="apikey_value_cannot_be_empty"))
        elif('token' in self.wml_credentials.keys()):
            if (self.wml_credentials['token'] != ''):
                return True
            else:
                raise WMLClientError(Messages.get_message(message_id="token_value_cannot_be_empty"))
        else:
            return False



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import json,ast
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepository import PipelineArtifact
from ibm_watson_machine_learning.libs.repo.swagger_client.models.ml_assets_create_experiment_output import MlAssetsCreateExperimentOutput
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact import SparkPipelineLoader, SparkPipelineContentLoader,\
    IBMSparkPipelineContentLoader, MLPipelineContentLoader
from ibm_watson_machine_learning.libs.repo.util import Json2ObjectMapper
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.generic_file_pipeline_model_loader import GenericFilePipelineModelLoader
from ibm_watson_machine_learning.libs.repo.mlrepository.generic_archive_pipeline_artifact import GenericArchivePipelineArtifact
from ibm_watson_machine_learning.libs.repo.util.generic_archive_file_check import GenericArchiveFrameworkCheck
lib_checker = LibraryChecker()

class ExperimentAdapter(object):
    """
    Adapter creating pipeline artifact using output from service.
    """
    @staticmethod
    def __strip_output(output):
        return ast.literal_eval(json.dumps(output))

    def __init__(self, pipeline_output, version_output, client):
        self.pipeline_output = pipeline_output
        self.version_output = version_output
        self.client = client
        self.pipeline_entity = self.__strip_output(pipeline_output.entity)
        if pipeline_output is not None and not isinstance(pipeline_output, MlAssetsCreateExperimentOutput):
            raise ValueError('Invalid type for pipeline: {}'.format(pipeline_output.__class__.__name__))

        #if self.pipeline_entity['framework'] is not None:
        #    self.pipeline_type = pipeline_output.entity['framework']


    def artifact(self):
#       if re.match('sparkml-pipeline-\d+\.\d+', self.pipeline_type['name']) is not None:
#         if re.match('mllib', self.pipeline_type['name']) is not None:
#             lib_checker.check_lib(PYSPARK)
#             pipeline_artifact_builder = type(
#                 "PipelineArtifact",
#                 (SparkPipelineContentLoader, SparkPipelineLoader, PipelineArtifact, object),
#                 {}
#             )
#         elif re.match('wml', self.pipeline_type['name']) is not None:
#             lib_checker.check_lib(PYSPARK)
#             lib_checker.check_lib(MLPIPELINE)
#             pipeline_artifact_builder = type(
#                 "WMLPipelineArtifact",
#                 (MLPipelineContentLoader, SparkPipelineLoader, PipelineArtifact, object),
#                 {}
#             )
#         elif GenericArchiveFrameworkCheck.is_archive_framework(self.pipeline_type['name']):
#             pipeline_artifact_builder = type(
#                 "GenericArchivePipelineArtifact",
#                 (GenericFilePipelineModelLoader, GenericArchivePipelineArtifact, object),
#                 {}
#             )
#         else:
#             raise ValueError('Invalid pipeline_type: {}'.format(self.pipeline_type['name']))

        prop_map = {
            MetaNames.CREATION_TIME: self.pipeline_output.metadata.created_at,
            MetaNames.LAST_UPDATED: self.pipeline_output.metadata.modified_at,
            #MetaNames.FRAMEWORK_VERSION: self.pipeline_type['version'],
            MetaNames.NAME: self.pipeline_output.entity.name
        }
        #if 'libraries' in self.pipeline_type:
        #    if self.pipeline_type.get('libraries') is not None:
        #        prop_map[MetaNames.FRAMEWORK_LIBRARIES] = self.pipeline_type.get('libraries')

        if self.pipeline_entity.get('description', None) is not None:
            prop_map[MetaNames.DESCRIPTION] = self.pipeline_entity['description']

        #if self.pipeline_entity.get('author', None) is not None:
        #    authorval = self.pipeline_entity.get('author')
        #    if authorval.get('name', None) is not None:
        #        prop_map[MetaNames.AUTHOR_NAME] = authorval['name']

        if self.pipeline_entity.get('tags', None) is not None:
            prop_map[MetaNames.TAGS] = self.pipeline_entity.get('tags')

        if self.pipeline_entity.get('import', None) is not None:
            # prop_map[MetaNames.TRAINING_DATA_REF] = str(self.pipeline_entity['training_data']).encode('ascii')
            prop_map[MetaNames.IMPORT] = str(self.pipeline_entity['import'])
        if self.pipeline_entity.get('document', None) is not None:
            # prop_map[MetaNames.TRAINING_DATA_REF] = str(self.pipeline_entity['training_data']).encode('ascii')
            prop_map[MetaNames.DOCUMENT] = str(self.pipeline_entity['document'])
      #  name = ''.join(map(lambda x: chr(ord(x)),self.pipeline_output.entity['name']))
        name =  self.pipeline_output.entity.name

#        pipeline_url = self.pipeline_output.metadata.url
        pipeline_url = self.pipeline_output.metadata.href

        pipeline_id = pipeline_url.split("/pipelines/")[1].split("/")[0]

        # pipeline_artifact = pipeline_artifact_builder(
        #     pipeline_id,
        #     name,
        #     MetaProps(prop_map))

        #pipeline_artifact.client = self.client

        #if self.version_output is not None:
        #    version_url = self.version_output['href']
        #    training_definition_url = version_url.split("/versions")[0]



#            version_prop_map = {
#                MetaNames.VERSION: self.version_output['guid'],
#                MetaNames.TRAINING_DEFINITION_VERSION_URL:version_url,
#                MetaNames.TRAINING_DEFINITION_URL:training_definition_url
#            }
#            pipeline_artifact._content_href = self.version_output['content_url']

#        else:
#            version_prop_map = {
#                MetaNames.VERSION: self.pipeline_output.entity['training_definition_version']['guid']
#            }
#            pipeline_artifact._content_href = self.pipeline_output.entity['training_definition_version']['content_url']

#        pipeline_artifact.meta.merge(MetaProps(version_prop_map))

        return pipeline_artifact



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on Feb 17, 2017

@author: calin
'''
import logging

from pyspark.ml.param import Params
from pyspark.ml.pipeline import Pipeline
from pyspark.ml.wrapper import JavaEstimator
from pyspark.ml.wrapper import _jvm

__all__ = ['Learner', 'Target']

logger = logging.getLogger("mlpipelinepy")

def estimatorCopy(estimator, jvmObject, extra=None):
    if extra is None:
        extra = dict()
    if isinstance(estimator , Pipeline):
        that = Params.copy(estimator, extra)
        stages = [estimatorCopy(stage, jvmObject, extra) for stage in that.getStages()]
        return that.setStages(stages)
    elif (isinstance(estimator,JavaEstimator)):
        newJavaEstimator = estimator.copy(extra)
        paramMapConstructor = jvmObject.org.apache.spark.ml.param.ParamMap
        jParamMap = paramMapConstructor()
        newJavaEstimator._java_obj = newJavaEstimator._java_obj.copy(jParamMap)
        return newJavaEstimator
    else :
        return estimator.copy(extra)


def emptyParam():
    jvmObject = _jvm()
    return jvmObject.org.apache.spark.ml.param.ParamMap.empty()


import abc
from os import environ

import unittest

from sklearn.pipeline import Pipeline

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService, Batch
from ibm_watson_machine_learning.workspace import WorkSpace
from ibm_watson_machine_learning.helpers.connections import DataConnection
from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.tests.utils import (get_wml_credentials, get_cos_credentials, get_space_id,
                                                     is_cp4d)
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup, delete_model_deployment
from ibm_watson_machine_learning.utils.autoai.enums import TShirtSize, PredictionType

from ibm_watson_machine_learning.tests.utils.assertions import get_and_predict_all_pipelines_as_lale


class AbstractTestWebservice(abc.ABC):
    """
    The abstract tests which covers:
    - deployment with lale pipeline
    - deployment deletion
    In order to execute test connection definitions must be provided
    in inheriting classes.
    """
    SPACE_ONLY = True

    DEPLOYMENT_NAME = "SDK tests Deployment"

    wml_client: 'APIClient' = None
    remote_auto_pipelines: 'RemoteAutoPipelines' = None
    wml_credentials = None
    service: 'WebService' = None

    space_id = None
    project_id = None
    target_space_id = None

    X_df = None

    @abc.abstractmethod
    def test_00a_space_cleanup(self):
        pass

    ###########################################################
    #      DEPLOYMENT SECTION    tests numbers start from 31  #
    ###########################################################

    def test_31_deployment_setup_and_preparation(self):
        # note: if target_space_id is not set, use the space_id
        if self.target_space_id is None:
            self.target_space_id = self.space_id
        # end note

        if self.SPACE_ONLY:
            AbstractTestWebservice.service = WebService(source_wml_credentials=self.wml_credentials,
                                                        source_space_id=self.space_id,
                                                        target_wml_credentials=self.wml_credentials,
                                                        target_space_id=self.target_space_id)
        else:
            AbstractTestWebservice.service = WebService(source_wml_credentials=self.wml_credentials,
                                                        source_project_id=self.project_id,
                                                        target_wml_credentials=self.wml_credentials,
                                                        target_space_id=self.target_space_id)

        self.wml_client.set.default_space(self.space_id)
        delete_model_deployment(self.wml_client, deployment_name=self.DEPLOYMENT_NAME)

        self.assertIsInstance(AbstractTestWebservice.service, WebService, msg="Deployment is not of WebService type.")
        self.assertIsInstance(AbstractTestWebservice.service._source_workspace, WorkSpace, msg="Workspace set incorrectly.")
        self.assertEqual(AbstractTestWebservice.service.id, None, msg="Deployment ID initialized incorrectly")
        self.assertEqual(AbstractTestWebservice.service.name, None, msg="Deployment name initialized incorrectly")

    def test_32__deploy__deploy_best_computed_pipeline_from_autoai_on_wml(self):
        best_pipeline = self.remote_auto_pipelines.summary()._series['Enhancements'].keys()[0]
        print('Deploying', best_pipeline)
        AbstractTestWebservice.service.create(
            experiment_run_id=self.remote_auto_pipelines._engine._current_run_id,
            model=best_pipeline,
            deployment_name=self.DEPLOYMENT_NAME)

        self.assertIsNotNone(AbstractTestWebservice.service.id, msg="Online Deployment creation - missing id")
        self.assertIsNotNone(AbstractTestWebservice.service.name, msg="Online Deployment creation - name not set")
        self.assertIsNotNone(AbstractTestWebservice.service.scoring_url,
                             msg="Online Deployment creation - mscoring url  missing")

    def test_33_score_deployed_model(self):
        nb_records = 5
        predictions = AbstractTestWebservice.service.score(payload=self.X_df[:nb_records])
        print(predictions)
        self.assertIsNotNone(predictions)
        self.assertEqual(len(predictions['predictions'][0]['values']), nb_records)

    def test_34_list_deployments(self):
        AbstractTestWebservice.service.list()
        params = AbstractTestWebservice.service.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_35_delete_deployment(self):
        print("Delete current deployment: {}".format(AbstractTestWebservice.service.deployment_id))
        AbstractTestWebservice.service.delete()
        self.wml_client.set.default_space(self.space_id) if not self.wml_client.default_space_id else None
        self.wml_client.repository.delete(AbstractTestWebservice.service.asset_id)
        self.wml_client.set.default_project(self.project_id) if is_cp4d() else None
        self.assertEqual(AbstractTestWebservice.service.id, None, msg="Deployment ID deleted incorrectly")
        self.assertEqual(AbstractTestWebservice.service.name, None, msg="Deployment name deleted incorrectly")
        self.assertEqual(AbstractTestWebservice.service.scoring_url, None,
                         msg="Deployment scoring_url deleted incorrectly")



#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.utils.autoai.utils import check_dependencies_versions
from ibm_watson_machine_learning.tests.utils import get_wml_credentials


class MyTestCase(unittest.TestCase):
    request_json = {'hybrid_pipeline_software_specs': [{'name': "autoai-kb_3.1-py3.7"}]}

    @classmethod
    def setUpClass(cls) -> None:
        """
        Load WML credentials from config.ini file based on ENV variable.
        """

        cls.wml_credentials = get_wml_credentials()
        cls.wml_client = APIClient(wml_credentials=cls.wml_credentials.copy())

    def test_01__all_and_xgboost(self):
        check_dependencies_versions(self.request_json, self.wml_client, estimator_pkg='xgboost')

    def test_02__all_and_lightgbm(self):
        check_dependencies_versions(self.request_json, self.wml_client, estimator_pkg='lightgbm')


if __name__ == '__main__':
    unittest.main()



import unittest,time

import logging
from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *
from ibm_watson_machine_learning.tests.ICP.models_preparation import *


class TestRshinyApp(unittest.TestCase):
    runtime_uid = None
    deployment_uid = None
    function_uid = None
    scoring_url = None
    logger = logging.getLogger(__name__)
    skip_tests = False

    @classmethod
    def setUpClass(self):
        TestRshinyApp.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "test_case_Rshiny_app" + time.asctime() })
        TestRshinyApp.space_id = self.client.spaces.get_uid(self.space)
        print('space_id: ', TestRshinyApp.space_id)
        self.client.set.default_space(TestRshinyApp.space_id)

        if self.wml_credentials['version'] == '2.0':
           TestRshinyApp.skip_tests = True

    # def test_01_service_instance_details(self):
    #     TestAIFunction.logger.info("Check client ...")
    #     self.assertTrue(self.client.__class__.__name__ == 'APIClient')
    #
    #     TestAIFunction.logger.info("Getting instance details ...")
    #     details = self.client.service_instance.get_details()
    #     TestAIFunction.logger.debug(details)
    #
    #     self.assertTrue("published_models" in str(details))
    #     self.assertEqual(type(details), dict)

    def test_01_create_shiny_asset(self):
        if not TestRshinyApp.skip_tests:

            self.client.shiny.ConfigurationMetaNames.show()


            meta_prop_shiny = {
                self.client.shiny.ConfigurationMetaNames.NAME: "my shiny app",
                self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "shiny app for deployment"
            }

            shiny_details = self.client.shiny.store(meta_prop_shiny, file_path="artifacts/app.R.zip")

            TestRshinyApp.shiny_asset_uid = self.client.shiny.get_uid(shiny_details)
            TestRshinyApp.shiny_asset_url = self.client.shiny.get_href(shiny_details)
            TestRshinyApp.logger.info("shiny asset ID:" + str(TestRshinyApp.shiny_asset_uid))
            TestRshinyApp.logger.info("shiny asset URL:" + str(TestRshinyApp.shiny_asset_url))
            self.assertIsNotNone(TestRshinyApp.shiny_asset_uid)
            self.assertIsNotNone(TestRshinyApp.shiny_asset_url)


    def test_02_download_shiny_content(self):
        if not TestRshinyApp.skip_tests:
            try:
                os.remove('test_shiny_asset.zip')
            except:
                pass
            self.client.shiny.download(TestRshinyApp.shiny_asset_uid, filename='test_shiny_asset.zip')
            try:
                os.remove('test_shiny_asset.zip')
            except:
                pass

    def test_03_get_details(self):
        if not TestRshinyApp.skip_tests:

            details = self.client.shiny.get_details(TestRshinyApp.shiny_asset_uid)
            self.assertTrue(TestRshinyApp.shiny_asset_uid in str(details))

    def test_04_list(self):
        if not TestRshinyApp.skip_tests:
            self.client.shiny.list()

    def test_05_revisions(self):
        if not TestRshinyApp.skip_tests:
            meta_prop_shiny_revision1 = {
                self.client.shiny.ConfigurationMetaNames.NAME: "Shiny revision 1",
                self.client.shiny.ConfigurationMetaNames.DESCRIPTION: "Shiny revision 1"
            }

            # Update meta and attachment for revision 1 creation
            shiny_update = self.client.shiny.update(TestRshinyApp.shiny_asset_uid,
                                                    meta_prop_shiny_revision1,
                                                    file_path="artifacts/app1.R.zip")
            attachment_id = shiny_update[u'metadata'][u'attachment_id']

            # Create revision
            revision = self.client.shiny.create_revision(TestRshinyApp.shiny_asset_uid)

            self.assertTrue(revision[u'metadata'][u'revision_id'] and revision[u'metadata'][u'revision_id'] == 1)

            new_attachment_id = revision[u'metadata'][u'attachment_id']

            self.assertTrue(attachment_id != new_attachment_id)

            # List revisions
            self.client.shiny.list_revisions(TestRshinyApp.shiny_asset_uid)

            try:
                os.remove('test_shiny_asset_rev1.zip')
            except:
                pass

            # Download revision 1 attachment
            self.client.shiny.download(TestRshinyApp.shiny_asset_uid, filename='test_shiny_asset_rev1.zip', rev_uid=1)
            try:
                os.remove('test_shiny_asset_rev1.zip')
            except:
                pass

    def test_06_create_deployment(self):
        if not TestRshinyApp.skip_tests:
            deploy_meta = {
                    self.client.deployments.ConfigurationMetaNames.NAME: "deployment_rshiny",
                    self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment rshiny deployment",
                    self.client.deployments.ConfigurationMetaNames.R_SHINY: {"authentication" : "anyone_with_url" },
                    self.client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {"name":"S", "num_nodes":1}
                }

            TestRshinyApp.logger.info("Create deployment")
            deployment = self.client.deployments.create(artifact_uid=TestRshinyApp.shiny_asset_uid, meta_props=deploy_meta)
            TestRshinyApp.logger.debug("deployment: " + str(deployment))
            # TestRshinyApp.scoring_url = self.client.deployments.get_scoring_href(deployment)
            # TestRshinyApp.logger.debug("Scoring href: {}".format(TestRshinyApp.scoring_url))
            TestRshinyApp.deployment_uid = self.client.deployments.get_uid(deployment)
            TestRshinyApp.logger.debug("Deployment uid: {}".format(TestRshinyApp.deployment_uid))
            self.client.deployments.list()
            self.assertTrue("deployment_rshiny" in str(deployment))

    def test_07_update_deployment(self):
        if not TestRshinyApp.skip_tests:
            patch_meta = {
                self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment_Updated_Shiny_Description",
            }
            self.client.deployments.update(TestRshinyApp.deployment_uid, patch_meta)

    def test_08_get_deployment_details(self):
        if not TestRshinyApp.skip_tests:
            TestRshinyApp.logger.info("Get deployment details")
            deployment_details = self.client.deployments.get_details()
            TestRshinyApp.logger.debug("Deployment details: {}".format(deployment_details))
            self.assertTrue('deployment_Updated_Shiny_Description' in str(deployment_details))

    def test_09_delete_deployment(self):
        if not TestRshinyApp.skip_tests:
            TestRshinyApp.logger.info("Delete deployment")
            self.client.deployments.delete(TestRshinyApp.deployment_uid)

    def test_10_delete_shiny_asset(self):
        if not TestRshinyApp.skip_tests:
            TestRshinyApp.logger.info("Delete function")
            self.client.shiny.delete(TestRshinyApp.shiny_asset_uid)
            self.client.spaces.delete(TestRshinyApp.space_id)


if __name__ == '__main__':
    unittest.main()



import abc
import os
from datetime import datetime
from os import environ

import unittest

from sklearn.pipeline import Pipeline

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.deployment import Batch
from ibm_watson_machine_learning.workspace import WorkSpace
from ibm_watson_machine_learning.experiment.autoai.optimizers import RemoteAutoPipelines
from ibm_watson_machine_learning.helpers.connections import DataConnection, AssetLocation, \
    DeploymentOutputAssetLocation, S3Location
from ibm_watson_machine_learning.tests.utils import save_multiple_data_to_cos_bucket, save_data_to_cos_bucket
from ibm_watson_machine_learning.tests.utils.cleanup import space_cleanup, delete_model_deployment
from ibm_watson_machine_learning.utils.autoai.enums import TShirtSize, PredictionType


class AbstractTestBatch(abc.ABC):
    """
    The abstract tests which covers:
    - deployment with lale pipeline
    - deployment deletion
    In order to execute test connection definitions must be provided
    in inheriting classes.
    """

    # FLAGS
    SPACE_ONLY: bool  # if True, experiment was run on space.

    BATCH_DEPLOYMENT_WITH_DF = True  # batch input passed as Pandas.DataFrame
    BATCH_DEPLOYMENT_WITH_DA = True  # batch input passed as DataConnection type data-assets with csv files
    BATCH_DEPLOYMENT_WITH_CDA = True  # batch input passed as DataConnection type data-assets with connection_id(COS)
    BATCH_DEPLOYMENT_WITH_CA = True  # batch input passed as DataConnection type connected assets (COS)
    # end FLAGS

    DEPLOYMENT_NAME = "SDK tests Deployment"

    wml_client: 'APIClient'
    remote_auto_pipelines: 'RemoteAutoPipelines'
    wml_credentials: str
    service_batch: 'Batch'

    asset_id: str
    batch_payload_location: str
    batch_output_filename = f"batch_output_{datetime.utcnow().isoformat()}.csv"

    space_id: str
    project_id: str

    data_assets_to_delete: list = []

    X_df = None

    @abc.abstractmethod
    def test_00a_space_cleanup(self):
        pass

    ########################################################
    #  Batch deployment (possible test numbers are: 40-54) #
    ########################################################

    def test_40_deployment_target_space_setup(self):
        # note: if target_space_id is not set, use the space_id
        if self.target_space_id is None:
            AbstractTestBatch.target_space_id = self.space_id
        else:
            AbstractTestBatch.target_space_id = self.target_space_id
        # end note

        self.wml_client.set.default_space(AbstractTestBatch.target_space_id)

    def test_41_batch_deployment_setup_and_preparation(self):

        self.assertIsNotNone(AbstractTestBatch.target_space_id, "Test issue: target space not set.")

        if self.SPACE_ONLY:
            AbstractTestBatch.service_batch = Batch(source_wml_credentials=self.wml_credentials,
                                                    source_space_id=self.space_id,
                                                    target_wml_credentials=self.wml_credentials,
                                                    target_space_id=AbstractTestBatch.target_space_id)
        else:
            AbstractTestBatch.service_batch = Batch(source_wml_credentials=self.wml_credentials,
                                                    source_project_id=self.project_id,
                                                    target_wml_credentials=self.wml_credentials,
                                                    target_space_id=AbstractTestBatch.target_space_id)

        self.assertIsInstance(AbstractTestBatch.service_batch, Batch, msg="Deployment is not of Batch type.")
        self.assertIsInstance(AbstractTestBatch.service_batch._source_workspace, WorkSpace,
                              msg="Workspace set incorrectly.")
        self.assertEqual(AbstractTestBatch.service_batch.id, None, msg="Deployment ID initialized incorrectly")
        self.assertEqual(AbstractTestBatch.service_batch.name, None, msg="Deployment name initialized incorrectly")

    def test_42_deploy__batch_deploy_best_computed_pipeline_from_autoai_on_wml(self):
        AbstractTestBatch.service_batch.create(
            experiment_run_id=self.remote_auto_pipelines._engine._current_run_id,
            model="Pipeline_2",
            deployment_name=self.DEPLOYMENT_NAME + ' BATCH')

        self.assertIsNotNone(AbstractTestBatch.service_batch.id, msg="Batch Deployment creation - missing id")
        self.assertIsNotNone(AbstractTestBatch.service_batch.id, msg="Batch Deployment creation - name not set")
        self.assertIsNotNone(AbstractTestBatch.service_batch.asset_id,
                             msg="Batch Deployment creation - model (asset) id missing, incorrect model storing")

    def test_43_list_batch_deployments(self):
        deployments = AbstractTestBatch.service_batch.list()
        print(deployments)
        self.assertIn('created_at', str(deployments).lower())
        self.assertIn('status', str(deployments).lower())

        params = AbstractTestBatch.service_batch.get_params()
        print(params)
        self.assertIsNotNone(params)

    def test_44_run_job__batch_deployed_model_with_data_frame(self):
        if self.BATCH_DEPLOYMENT_WITH_DF:
            scoring_params = AbstractTestBatch.service_batch.run_job(
                payload=self.X_df,
                background_mode=False)
            print(scoring_params)
            print(AbstractTestBatch.service_batch.get_job_result(scoring_params['metadata']['id']))
            self.assertIsNotNone(scoring_params)
            self.assertIn('predictions', str(scoring_params).lower())
        else:
            self.skipTest("Skip batch deployment run job with data frame")

    def test_45_run_job_batch_deployed_model_with_data_assets(self):
        data_connections_space_only = []
        # self.wml_client.set.default_space(self.target_space_id)

        test_case_batch_output_filename = "da_" + self.batch_output_filename

        results_reference = DataConnection(
            location=DeploymentOutputAssetLocation(name=test_case_batch_output_filename))

        if not self.BATCH_DEPLOYMENT_WITH_DA:
            self.skipTest("Skip batch deployment run job with data asset type")
        else:
            self.assertIsNotNone(self.batch_payload_location,
                                 "Test configuration failure: Batch payload location is missing")

            asset_details = self.wml_client.data_assets.create(
                name=self.batch_payload_location.split('/')[-1],
                file_path=self.batch_payload_location)
            asset_id = self.wml_client.data_assets.get_id(asset_details)
            AbstractTestBatch.data_assets_to_delete.append(asset_id)
            data_connections_space_only = [DataConnection(location=AssetLocation(asset_id=asset_id))]

        scoring_params = AbstractTestBatch.service_batch.run_job(
            payload=data_connections_space_only,
            output_data_reference=results_reference,
            background_mode=False)

        print(scoring_params)
        self.assertIsNotNone(scoring_params)

        deployment_job_id = self.wml_client.deployments.get_job_uid(scoring_params)

        self.wml_client.data_assets.list()

        data_asset_details = self.wml_client.data_assets.get_details()
        self.assertIn(test_case_batch_output_filename, str(data_asset_details),
                      f"Batch output file: {test_case_batch_output_filename} missed in data assets")

        predictions = AbstractTestBatch.service_batch.get_job_result(deployment_job_id)
        print(predictions)

        self.assertIsNotNone(predictions)

    def test_46_run_job_batch_deployed_model_with_data_assets_with_cos_connection(self):
        if not self.BATCH_DEPLOYMENT_WITH_CDA:
            self.skipTest("Skip batch deployment run job with data asset with cos connection type")

        data_connections_space_only = []
        # self.wml_client.set.default_space(self.target_space_id)

        test_case_batch_output_filename = "cos_da_" + self.batch_output_filename

        results_reference = DataConnection(
            location=DeploymentOutputAssetLocation(name=test_case_batch_output_filename))

        # prepare connection
        connection_details = self.wml_client.connections.create({
            'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name(
                'bluemixcloudobjectstorage'),
            'name': 'Connection to COS for tests',
            'properties': {
                'bucket': self.bucket_name,
                'access_key': self.cos_credentials['cos_hmac_keys']['access_key_id'],
                'secret_key': self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                'iam_url': self.wml_client.service_instance._href_definitions.get_iam_token_url(),
                'url': self.cos_endpoint
            }
        })

        AbstractTestBatch.connection_id = self.wml_client.connections.get_uid(connection_details)

        self.assertIsNotNone(self.batch_payload_location,
                             "Test configuration failure: Batch payload location is missing")

        AbstractTestBatch.bucket_name = save_data_to_cos_bucket(self.batch_payload_location,
                                                                self.batch_payload_cos_location,
                                                                access_key_id=self.cos_credentials['cos_hmac_keys'][
                                                                    'access_key_id'],
                                                                secret_access_key=
                                                                self.cos_credentials['cos_hmac_keys'][
                                                                    'secret_access_key'],
                                                                cos_endpoint=self.cos_endpoint,
                                                                bucket_name=self.bucket_name)

        asset_details = self.wml_client.data_assets.store({
            self.wml_client.data_assets.ConfigurationMetaNames.CONNECTION_ID: AbstractTestBatch.connection_id,
            self.wml_client.data_assets.ConfigurationMetaNames.NAME: "Batch deployment asset",
            self.wml_client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: self.batch_payload_cos_location
        })

        asset_id = self.wml_client.data_assets.get_id(asset_details)
        self.assertIsInstance(asset_id, str)
        AbstractTestBatch.data_assets_to_delete.append(asset_id)

        data_connections_space_only.append(
            DataConnection(data_asset_id=asset_id))

        self.assertEqual(len(data_connections_space_only), 1)

        scoring_params = AbstractTestBatch.service_batch.run_job(
            payload=data_connections_space_only,
            output_data_reference=results_reference,
            background_mode=False)

        print(scoring_params)
        self.assertIsNotNone(scoring_params)

        deployment_job_id = self.wml_client.deployments.get_job_uid(scoring_params)

        self.wml_client.data_assets.list()

        data_asset_details = self.wml_client.data_assets.get_details()
        self.assertIn(test_case_batch_output_filename, str(data_asset_details),
                      f"Batch output file: {test_case_batch_output_filename} missed in data assets")

        predictions = AbstractTestBatch.service_batch.get_job_result(deployment_job_id)
        print(predictions)

        self.assertIsNotNone(predictions)

    def test_47_run_job_batch_deployed_model_with_connected_data_asset(self):

        if not self.BATCH_DEPLOYMENT_WITH_CA:
            self.skipTest("Skip batch deployment run job with connected asset with cos connection type")
        data_connections_space_only = []
        self.wml_client.set.default_space(self.target_space_id)

        test_case_batch_output_filename = "cos_ca_" + self.batch_output_filename

        results_reference = DataConnection(
            location=DeploymentOutputAssetLocation(name=test_case_batch_output_filename))

        # prepare connection
        if not AbstractTestBatch.connection_id:
            connection_details = self.wml_client.connections.create({
                'datasource_type': self.wml_client.connections.get_datasource_type_uid_by_name(
                    'bluemixcloudobjectstorage'),
                'name': 'Connection to COS for tests',
                'properties': {
                    'bucket': self.bucket_name,
                    'access_key': self.cos_credentials['cos_hmac_keys']['access_key_id'],
                    'secret_key': self.cos_credentials['cos_hmac_keys']['secret_access_key'],
                    'iam_url': self.wml_client.service_instance._href_definitions.get_iam_token_url(),
                    'url': self.cos_endpoint
                }
            })

            AbstractTestBatch.connection_id = self.wml_client.connections.get_uid(connection_details)


        self.assertIsNotNone(self.batch_payload_location,
                             "Test configuration failure: Batch payload location is missing")

        AbstractTestBatch.bucket_name = save_data_to_cos_bucket(self.batch_payload_location,
                                                                self.batch_payload_cos_location,
                                                                access_key_id=self.cos_credentials['cos_hmac_keys'][
                                                                    'access_key_id'],
                                                                secret_access_key=
                                                                self.cos_credentials['cos_hmac_keys'][
                                                                    'secret_access_key'],
                                                                cos_endpoint=self.cos_endpoint,
                                                                bucket_name=self.bucket_name)
        conn_space = DataConnection(
            connection_asset_id=AbstractTestBatch.connection_id,
            location=S3Location(
                bucket=self.bucket_name,
                path=self.data_cos_path
            )
        )

        data_connections_space_only.append(conn_space)

        scoring_params = AbstractTestBatch.service_batch.run_job(
            payload=data_connections_space_only,
            output_data_reference=results_reference,
            background_mode=False)

        print(scoring_params)
        self.assertIsNotNone(scoring_params)

        deployment_job_id = self.wml_client.deployments.get_job_uid(scoring_params)

        self.wml_client.data_assets.list()

        data_asset_details = self.wml_client.data_assets.get_details()
        self.assertIn(test_case_batch_output_filename, str(data_asset_details),
                      f"Batch output file: {test_case_batch_output_filename} missed in data assets")

        predictions = AbstractTestBatch.service_batch.get_job_result(deployment_job_id)

        print(predictions)
        self.assertIsNotNone(predictions)

    def test_48_run_job_batch_deployed_model_with_data_connection_container(self):
        if self.wml_client.ICP or self.wml_client.WSD:
            self.skipTest("Batch Deployment with container data connection is available only for Cloud")
        else:
            self.skipTest("not ready")

    def test_49_delete_deployment_batch(self):
        print("Delete current deployment: {}".format(AbstractTestBatch.service_batch.deployment_id))
        AbstractTestBatch.service_batch.delete()
        self.wml_client.set.default_space(self.target_space_id) if not self.wml_client.default_space_id else None
        self.wml_client.repository.delete(AbstractTestBatch.service_batch.asset_id)

        print(f"Delete all created assets in Batch deployment tests. {AbstractTestBatch.data_assets_to_delete}")
        for asset_id in AbstractTestBatch.data_assets_to_delete:
            self.wml_client.data_assets.delete(asset_id)

        self.assertEqual(AbstractTestBatch.service_batch.id, None, msg="Deployment ID deleted incorrectly")
        self.assertEqual(AbstractTestBatch.service_batch.name, None, msg="Deployment name deleted incorrectly")
        self.assertEqual(AbstractTestBatch.service_batch.scoring_url, None,
                         msg="Deployment scoring_url deleted incorrectly")



import unittest
import pandas as pd
from sklearn.pipeline import Pipeline
from lale.operators import TrainablePipeline

from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.utils.autoai.enums import (
    PredictionType, Metrics, ClassificationAlgorithms, PipelineTypes)
from ibm_watson_machine_learning.experiment.autoai.optimizers import LocalAutoPipelines


class TestLocalAutoPipelines(unittest.TestCase):
    data_location = './autoai/data/bank.csv'
    local_optimizer: 'LocalAutoPipelines' = None
    best_pipeline_model: 'Pipeline' = None
    pipeline = None

    @classmethod
    def setUp(cls) -> None:
        cls.data = pd.read_csv(cls.data_location)
        cls.X = cls.data.drop(['y'], axis=1)
        cls.y = cls.data['y']

    def test_01_initialize_LocalAutoPipelines__LocalAutoPipelines_initialized(self):
        TestLocalAutoPipelines.local_optimizer = AutoAI().optimizer(
            name='Local optimization',
            desc='Test local optimization',
            prediction_type=PredictionType.BINARY,
            prediction_column='state_code',
            scoring=Metrics.ROC_AUC_SCORE,
            test_size=0.2,
            max_number_of_estimators=1,
            daub_include_only_estimators=[
                ClassificationAlgorithms.LGBM,
                ClassificationAlgorithms.DT,
                ClassificationAlgorithms.XGB,
                ClassificationAlgorithms.GB,
                ClassificationAlgorithms.EX_TREES
            ]
        )

        self.assertIsInstance(self.local_optimizer, LocalAutoPipelines, msg="Local Pipeline Optimizer not initialized")

    def test_02__fit__start_local_training__local_training_finished_and_pipeline_model_returned(self):
        TestLocalAutoPipelines.best_pipeline_model = self.local_optimizer.fit(X=self.X, y=self.y)

        self.assertIsInstance(self.best_pipeline_model, Pipeline, msg="Computed best pipeline is not of sklearn type.")

    def test_03__get_params__get_configuration_parameters_of_local_auto_pipeline__parameters_listed(self):
        parameters = self.local_optimizer.get_params()
        print(parameters)
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_04_predict__predict_using_fitted_pipeline__predictions_are_computed(self):
        predictions = self.local_optimizer.predict(X=self.X.values)
        print(predictions)

    def test_05__results__listing_all_pipelines__all_pipelines_listed_with_metrics(self):
        pipelines_details = self.local_optimizer.summary()
        print(pipelines_details)

    def test_06__get_pipeline_params__list_specific_pipeline_parameters__parameters_fetched_as_dict(self):
        pipeline_params = self.local_optimizer.get_pipeline_details(pipeline_name='Pipeline_1')
        print(pipeline_params)

    def test_07__get_pipeline_params__fetch_best_pipeline_parameters__parameters_fetched_as_dict(self):
        best_pipeline_params = self.local_optimizer.get_pipeline_details()
        print(best_pipeline_params)

    def test_08__get_pipeline__load_specified_pipeline__pipeline_loaded(self):
        TestLocalAutoPipelines.pipeline = self.local_optimizer.get_pipeline(pipeline_name='Pipeline_3')
        print(f"Fetched pipeline type: {type(self.pipeline)}")

        self.assertIsInstance(self.pipeline, TrainablePipeline,
                              msg="Fetched pipeline is not of TrainablePipeline instance.")

    def test_09__predict__do_the_predict_on_lale_pipeline__results_computed(self):
        predictions = self.pipeline.predict(self.X.values)
        print(predictions)

    def test_10__get_pipeline__load_specified_pipeline__sklearn_pipeline_loaded(self):
        TestLocalAutoPipelines.pipeline = self.local_optimizer.get_pipeline(
            pipeline_name='Pipeline_3', astype=PipelineTypes.SKLEARN)
        print(f"Fetched pipeline type: {type(self.pipeline)}")

        self.assertIsInstance(self.pipeline, Pipeline, msg="Fetched pipeline is not of sklearn type.")


if __name__ == '__main__':
    unittest.main()



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
import os

with open('VERSION', 'r') as f_ver:
    VERSION = f_ver.read()


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


# read the contents of README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='ibm_watson_machine_learning',
    version=VERSION,
    python_requires='>=3.7',
    author='IBM',
    author_email='svagaral@in.ibm.com, kaganesa@in.ibm.com, vbmithun@in.ibm.com, amadeusz.masny1@ibm.com',
    description='IBM Watson Machine Learning API Client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD',
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Natural Language :: English',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Microsoft :: Windows',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Information Technology',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Scientific/Engineering :: Information Analysis',
                 'Topic :: Internet'],
    keywords=["watson", "machine learning", "IBM", "Bluemix", "client", "API", "IBM Cloud"],
    url='http://ibm-wml-api-pyclient.mybluemix.net',
    packages=find_packages(),
    package_data={'': ['messages/messages_en.json']},
    install_requires=[
        'requests',
        'urllib3',
        'pandas<1.3.0,>=0.24.2',
        'certifi',
        'lomond',
        'tabulate',
        'packaging',
        'ibm-cos-sdk==2.7.*'
    ],
    include_package_data=True
)



#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import shutil

from ibm_watson_machine_learning.libs.repo.mlrepository.artifact_reader import ArtifactReader
from ibm_watson_machine_learning.libs.repo.util.compression_util import CompressionUtil
from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate


# TODO this part needs to be implemented
class SparkPipelineReader(ArtifactReader):
    def __init__(self, ml_pipeline, type_name):
        self.archive_path = None
        self.ml_pipeline = ml_pipeline
        self.type_name = type_name
        self.hummingbird_env = os.getenv('HB_RUNTIME_PROVIDER', "").upper()

    def read(self):
        return self._open_stream()

    def close(self):
        os.remove(self.archive_path)
        self.archive_path = None

    def _save_pipeline_archive(self):

        id_length = 20
        gen_id = uid_generate(id_length)
        temp_dir_name = '{}'.format(self.type_name + gen_id)
        if (self.hummingbird_env == 'HUMMINGBIRD') is True:
            temp_dir = os.path.join('/home/spark/shared/wml/repo/extract_', temp_dir_name)
        else:
            temp_dir = os.path.join('.', temp_dir_name)
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)

        self.ml_pipeline.write().overwrite().save(temp_dir)
        archive_path = self._compress_artifact(temp_dir, gen_id)
        shutil.rmtree(temp_dir)
        return archive_path

    def _compress_artifact(self, compress_artifact, gen_id):
        tar_filename = '{}_content.tar'.format(self.type_name + gen_id)
        gz_filename = '{}.gz'.format(tar_filename)
        CompressionUtil.create_tar(compress_artifact, '.', tar_filename)
        CompressionUtil.compress_file_gzip(tar_filename, gz_filename)
        os.remove(tar_filename)
        return gz_filename

    def _open_stream(self):
        if self.archive_path is None:
            self.archive_path = self._save_pipeline_archive()
        return open(self.archive_path, 'rb')



import unittest,time

import logging
from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *
from ibm_watson_machine_learning.tests.ICP.models_preparation import *
from ibm_watson_machine_learning.tests.ICP.preparation_and_cleaning import *


class TestConnections(unittest.TestCase):

    deployment_uid = None
    function_uid = None
    job_id = None
    scoring_url = None
    connection_asset_uid = None
    connected_data_asset_url = None
    connected_data_id = None
    connection_asset_url = None
    model_uid = None
    space_id = None
    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestConnections.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "test_case_scripts" + time.asctime() })
        TestConnections.space_id = self.client.spaces.get_id(self.space)
        self.client.set.default_space(TestConnections.space_id)


    def test_01_publish_model(self):
        # space = self.client.spaces.store({self.client.spaces.ConfigurationMetaNames.NAME: "test_case"})
        # space_id = self.client.spaces.get_uid(space)
        # self.client.set.default_space(space_id)
        # TestWMLClientWithScikitLearn.space_id = space_id
        # print("The space is" + space_id)
        TestConnections.logger.info("Creating scikit-learn model ...")

        model_data = create_scikit_learn_model_data()
        predicted = model_data['prediction']

        TestConnections.logger.debug(predicted)
        self.assertIsNotNone(predicted)

        self.logger.info("Publishing scikit-learn model ...")

        self.client.repository.ModelMetaNames.show()
        sw_spec_id = self.client.software_specifications.get_uid_by_name("scikit-learn_0.20-py3.6")
        model_props = {self.client.repository.ModelMetaNames.NAME: "ScikitModel",
                       self.client.repository.ModelMetaNames.TYPE: "scikit-learn_0.20",
                       self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_id
                       }
        published_model_details = self.client.repository.store_model(model=model_data['model'], meta_props=model_props, training_data=model_data['training_data'], training_target=model_data['training_target'])
        TestConnections.model_uid = self.client.repository.get_model_uid(published_model_details)
        TestConnections.model_url = self.client.repository.get_model_href(published_model_details)
        self.logger.info("Published model ID:" + str(TestConnections.model_uid))
        self.logger.info("Published model URL:" + str(TestConnections.model_url))
        self.assertIsNotNone(TestConnections.model_uid)
        self.assertIsNotNone(TestConnections.model_url)

    def test_02_create_connection_asset(self):
        import json
        self.client.connections.ConfigurationMetaNames.show()
        data_source_type_id = self.client.connections.get_datasource_type_uid_by_name('cloudobjectstorage')
        print(data_source_type_id)
        conn_properties = json.loads(config.get(environment, 'connection_cos_hmac_keys'))
        meta_props_connection = {
            self.client.connections.ConfigurationMetaNames.NAME: "Sample Connection from Python client",
            self.client.connections.ConfigurationMetaNames.DATASOURCE_TYPE: data_source_type_id,
            self.client.connections.ConfigurationMetaNames.DESCRIPTION: "Test connection object using fvt COS instance",
            self.client.connections.ConfigurationMetaNames.PROPERTIES: conn_properties
        }

        conn_details = self.client.connections.create(meta_props_connection)

        TestConnections.connection_asset_uid = self.client.connections.get_uid(conn_details)
        TestConnections.logger.info("Connection asset ID:" + str(TestConnections.connection_asset_uid))
        self.assertIsNotNone(TestConnections.connection_asset_uid)

    def test_03_create_connected_data(self):
        asset_meta_props = {
            self.client.data_assets.ConfigurationMetaNames.NAME: "scikit learn dataasset",
            self.client.data_assets.ConfigurationMetaNames.CONNECTION_ID: TestConnections.connection_asset_uid,
            self.client.data_assets.ConfigurationMetaNames.DESCRIPTION: "Test connection object using fvt COS instance",
            self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/wml-v4-fvt-remote-tests/heart_test.csv"
        }
        asset_details = self.client.data_assets.store(asset_meta_props)
        TestConnections.connected_data_asset_url = self.client.data_assets.get_href(asset_details)
        TestConnections.connected_data_id = self.client.data_assets.get_uid(asset_details)
        self.assertIsNotNone(TestConnections.connected_data_id)
        self.assertIsNotNone(TestConnections.connected_data_asset_url)

    def test_04_create_connection_mysql__asset(self):
        import json
        self.client.connections.ConfigurationMetaNames.show()
        mysql_data_source_type_id = self.client.connections.get_datasource_type_uid_by_name('mysql')
        print(mysql_data_source_type_id)
        conn_properties = json.loads(config.get(environment, 'mysql'))
        meta_props_connection = {
            self.client.connections.ConfigurationMetaNames.NAME: "Sample Connection for mysql from Python client",
            self.client.connections.ConfigurationMetaNames.DATASOURCE_TYPE: mysql_data_source_type_id,
            self.client.connections.ConfigurationMetaNames.DESCRIPTION: "Test connection object using mysql instance",
            self.client.connections.ConfigurationMetaNames.PROPERTIES: conn_properties
        }

        conn_details = self.client.connections.create(meta_props_connection)

        TestConnections.connection_mysql_asset_uid = self.client.connections.get_uid(conn_details)
        TestConnections.logger.info("Connection asset ID:" + str(TestConnections.connection_mysql_asset_uid))
        self.assertIsNotNone(TestConnections.connection_mysql_asset_uid)

    def test_05_create_connected_data(self):
        asset_meta_props = {
            self.client.data_assets.ConfigurationMetaNames.NAME: "mysql",
            self.client.data_assets.ConfigurationMetaNames.CONNECTION_ID: TestConnections.connection_mysql_asset_uid,
            self.client.data_assets.ConfigurationMetaNames.DESCRIPTION: "mysql",
            self.client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: "/wml-v4-fvt-remote-tests/test.csv"
        }
        asset_details = self.client.data_assets.store(asset_meta_props)
        print(asset_details)
        TestConnections.mysql_connected_data_asset_url = self.client.data_assets.get_href(asset_details)
        TestConnections.mysql_connected_data_id = self.client.data_assets.get_uid(asset_details)
        self.assertIsNotNone(TestConnections.mysql_connected_data_id)
        self.assertIsNotNone(TestConnections.mysql_connected_data_asset_url)

    def test_06_get_details(self):

        # details = self.client.connections.get_details(TestConnections.connection_asset_uid)
        # self.assertTrue(TestConnections.connection_asset_uid in str(details))

        details = self.client.connections.get_details(TestConnections.connection_mysql_asset_uid)
        print(details)
        self.assertTrue(TestConnections.connection_mysql_asset_uid in str(details))

    def test_07_list(self):
        self.client.connections.list()
        self.client.connections.list_datasource_types()

    def test_08_create_deployment(self):
        deploy_meta = {
                self.client.deployments.ConfigurationMetaNames.NAME: "deployment_using_connected_data",
                self.client.deployments.ConfigurationMetaNames.DESCRIPTION: "deployment rscript deployment",
                self.client.deployments.ConfigurationMetaNames.BATCH: {},
                self.client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {"name":"S", "num_nodes":1}
            }

        TestConnections.logger.info("Create deployment")
        deployment = self.client.deployments.create(artifact_uid=TestConnections.model_uid, meta_props=deploy_meta)
        TestConnections.logger.debug("deployment: " + str(deployment))
        # TestScripts.scoring_url = self.client.deployments.get_scoring_href(deployment)
        # TestScripts.logger.debug("Scoring href: {}".format(TestScripts.scoring_url))
        TestConnections.deployment_uid = self.client.deployments.get_uid(deployment)
        TestConnections.logger.debug("Deployment uid: {}".format(TestConnections.deployment_uid))
        self.client.deployments.list()
        self.assertTrue("deployment_using_connected_data" in str(deployment))

    def test_09_create_job(self):
        TestConnections.logger.info("Create job details")
        job_payload_ref = {
            self.client.deployments.ScoringMetaNames.INPUT_DATA_REFERENCES: [{
                "name": "test_ref_input",
                "type": "data_asset",
                "connection": {},
                "location": {
                    "href": TestConnections.connected_data_asset_url
                 }
            }],
            self.client.deployments.ScoringMetaNames.OUTPUT_DATA_REFERENCE: {

                "type": "data_asset",
                "connection": {},
                "location": {
                    "name": "scikit_{}.csv".format(TestConnections.deployment_uid),
                    "description": "testing csv results"
                }
            }
        }

        TestConnections.job_details = self.client.deployments.create_job(TestConnections.deployment_uid, meta_props=job_payload_ref)
        TestConnections.job_id = self.client.deployments.get_job_uid(TestConnections.job_details)

    def test_10_get_job_status(self):
        start_time = time.time()
        diff_time = start_time - start_time
        while True and diff_time < 10 * 60:
            time.sleep(3)
            response = self.client.deployments.get_job_status(TestConnections.job_id)
            if response['state'] == 'completed' or response['state'] == 'error' or response['state'] == 'failed':
                break
            diff_time = time.time() - start_time

        self.assertIsNotNone(response)
       # self.assertTrue(response['state'] == 'completed')

    def test_11_list_jobs(self):
        self.client.deployments.list_jobs()

    def test_12_delete_job(self):
        self.client.deployments.delete_job(TestConnections.job_id)

    def test_13_delete_deployment(self):
        TestConnections.logger.info("Delete deployment")
        self.client.deployments.delete(TestConnections.deployment_uid)

    def test_14_delete_connection_asset(self):
        TestConnections.logger.info("Delete connection and connected data")
        self.client.data_assets.delete(TestConnections.connected_data_id)
        self.client.connections.delete(TestConnections.connection_asset_uid)

    def test_14_delete_model_asset(self):
        TestConnections.logger.info("Delete model")
        self.client.repository.delete(TestConnections.model_uid)

    def test_15_delete_space(self):
        self.client.spaces.delete(TestConnections.space_id)


if __name__ == '__main__':
    unittest.main()




"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import abc
import time
import logging
import threading

from multiprocessing.pool import ThreadPool
from ibmfl.message.message import Message, ResponseMessage
from ibmfl.message.message_type import MessageType
from ibmfl.exceptions import LocalTrainingException
from ibmfl.party.status_type import StatusType

logger = logging.getLogger(__name__)


class PartyProtocolHandler(abc.ABC):
    """
    Base class for all PartyProtocolHandlers
    """

    def __init__(self, fl_model, connection, data_handler,
                 local_training_handler, metrics_recorder=None,
                 hyperparams=None, agg_info=None, synch=False, is_private=True,
                 **kwargs):
        """
        Initiate PartyProtocolHandler with provided fl_model, connection,
        data_handler and local_training_handler, hyperparams, aggregator
        information and synchronous flag are optional.

        :param fl_model: model to be trained
        :type fl_model: `model.FLModel`
        :param connection: connection that will be used to send messages
        :type connection: `Connection`
        :param data_handler: data handler that will be used to obtain data
        :type data_handler: `DataHandler`
        :param local_training_handler: local training handler class that \
        handles the access of the local model, for example, local training \
        of the model, prediction based on the model, etc.
        :type local_training_handler: `LocalTrainingHandler`
        :param hyperparams: Hyperparameters used for training.
        :type hyperparams: `Hyperparameters`
        :param agg_info: communication information related to aggregator
        :type agg_info: `dict`
        :param synch: get model update synchronously
        :type synch: `boolean`

        :return: None
        """

        self.fl_model = fl_model
        self.data_handler = data_handler
        self.connection = connection
        self.hyperparams = hyperparams
        self.agg_info = agg_info
        self.lock = threading.Lock()
        self.status = StatusType.IDLE
        self.synch = synch
        self.pool = ThreadPool(processes=1)
        self.local_training_handler = local_training_handler
        self.metrics_recorder = metrics_recorder
        if ( self.local_training_handler is not None ):
            self.local_training_handler.set_metrics_recorder_obj(self.metrics_recorder)
        self.is_private = is_private

    def get_handle(self, message_type):
        """
        Get handler for given message type.

        :param message_type: request message type
        :type message_type: `int`
        :return: a handler which was assigned for given message type
        """
        if message_type == MessageType.SAVE_MODEL.value:
            return self.local_training_handler.save_model
        elif message_type == MessageType.SYNC_MODEL.value:
            return self.local_training_handler.sync_model
        elif message_type == MessageType.EVAL_MODEL.value:
            return self.local_training_handler.eval_model
        elif message_type == MessageType.TRAIN.value:
            return self.local_training_handler.train
        else:
            raise LocalTrainingException("Unsupported message type!")

    def handle_request(self, msg):
        """
        Handle all incoming requests and route it to respective methods
        in local training handler.

        :param msg: Message object form connection
        :type msg: `Message`
        :return: Response message sent back to requester
        :rtype: ResponseMessage

        """
        logger.info("Received request from aggregator")
        message_type = msg.message_type
        logger.info("Received request in with message_type:  " +
                    str(message_type))

        data = msg.get_data()

        response_msg = ResponseMessage(req_msg=msg)
        response_data = {"status": "success"}
        status = True
        logger.info("Received request in PH " + str(message_type))

        try:
            if message_type is MessageType.TRAIN.value:
                if self.metrics_recorder:
                    self.metrics_recorder.add_entry()
                    self.metrics_recorder.set_round_no(self.local_training_handler.n_completed_trains)
            elif message_type is MessageType.STOP.value:
                self.status = StatusType.STOPPING
                # May be no need to send response
                response_msg = ResponseMessage(message_type=MessageType.ACK.value,
                                               id_request=-1,
                                               data={'ACK': True})
                logger.info("received a STOP request")
                return response_msg
        
            self.wait_for_model_initialization()
            handler = self.get_handle(message_type)

            response = handler(data.get('payload'))
            if message_type is MessageType.TRAIN.value:
                self.local_training_handler.n_completed_trains += 1
                if not self.is_private:
                    metrics_handler = self.get_handle(MessageType.EVAL_MODEL.value)
                    metrics = metrics_handler(data.get('payload'))
                    response_data['metrics'] = metrics
            elif message_type is MessageType.EVAL_MODEL.value:
                self.local_training_handler.n_completed_evals += 1

        except Exception as ex:
            logger.exception(ex)
            raise LocalTrainingException(
                "Error occurred while handling request")

        response_data['payload'] = response
        if not status:
            response_data['status'] = "error"

        response_msg.set_data(response_data)
        return response_msg

    def execute_async(self, id_request, msg):
        """
        Handle run in a different thread to allow asynchronous requests.

        :param msg: Message object form connection
        :type msg: `Message`
        """
        try:
            # Acquire lock so that we do not run train twice
            self.lock.acquire()
            logger.info("Handling async request in a separate thread")
            response_msg = self.handle_request(msg)

        except Exception as ex:
            logger.info("Exception occurred while async handling of msg: "
                        + msg)
            logger.exception(ex)

            response_msg = ResponseMessage(msg)
            response_msg.set_data({
                "status": "error",
                "payload": None
            })
        logger.info("successfully finished async request")
        self.connection.send_message(self.agg_info, response_msg)

        if self.metrics_recorder:
            with open(self.metrics_recorder.get_output_file(), 'w') as metrics_file:
                metrics_output_type = self.metrics_recorder.get_output_type()
                if metrics_output_type == 'json':
                    metrics_output_data = self.metrics_recorder.to_json()
                else:
                    logger.info("Bad metrics output filetype. Defaulting to json.")
                    metrics_output_data = self.metrics_recorder.to_json()
                metrics_file.write('{}\n'.format(metrics_output_data))

        # Release lock
        self.lock.release()
        return

    def handle_async_request(self, msg):
        """
        Handle all incoming requests asynchronously and route it to respective
         methods in local training handler.

        :param msg: Message object form connection
        :type msg: `Message`
        :return: Response message sent back to requester
        :rtype: ResponseMessage
        """
        try:
            response_msg = ResponseMessage(message_type=MessageType.ACK.value,
                                           id_request=-1,
                                           data={'ACK': True})
            logger.info("received a async request")

            id_request = msg.get_header()['id_request']

            self.pool.apply_async(
                self.execute_async, args=(id_request, msg))
            logger.info("finished async request")
        except Exception as ex:
            logger.info(ex)

        return response_msg

    def print_evaluate_local_model(self, hyperparams=None):
        """
        Print local evaluations in console

        :param hyperparams: hyperparams for evaluation on the local model
        :type hyperparams: `dict`
        :return: None
        """
        try:
            evaluations = self.local_training_handler.eval_model(hyperparams)
            # logger.info(evaluations)
        except Exception as ex:
            logger.info('Error occurred on evaluating model on local data. ' +
                        str(ex))


    def set_model(self, model):
        """Set model instance
        :param model: model to be trained
        :type model: `model.FLModel`
        :return: None
        """
        self.fl_model = model

    def set_training_handler(self, training_handler):
        """Set localtraining handler instance

        :param local_training_handler: local training handler class that \
        handles the access of the local model, for example, local training \
        of the model, prediction based on the model, etc.
        :type local_training_handler: `LocalTrainingHandler`        
        """
        self.local_training_handler = training_handler
        if ( self.local_training_handler is not None ):
            self.local_training_handler.set_metrics_recorder_obj(self.metrics_recorder)

    def wait_for_model_initialization(self):
        """Wait until model and localtraininghandler are initialized
        """
        logger.debug("Waiting for model initialization to finish")

        while not (self.fl_model and self.local_training_handler):
            time.sleep(10)


class PartyProtocolHandlerRabbitMQ(PartyProtocolHandler):
    """
    Extended class for PartyProtocolHandler for using with RabbitMQ connection
    """

    def __init__(
            self,
            fl_model,
            connection,
            data_handler,
            local_training_handler,
            hyperparams=None,
            agg_info=None,
            synch=True,
            is_private=True,
            **kwargs
    ):
        """
        Initiate PartyProtocolHandlerRabbitMQ with provided fl_model,
        connection and data_handler and hyperparams

        :param fl_model: model to be trained
        :type fl_model: `model.FLModel`
        :param connection: connection that will be used to send messages
        :type connection: `Connection`
        :param data_handler: data handler that will be used to obtain data
        :type data_handler: `DataHandler`
        :param hyperparams: Hyperparameters used for training.
        :type hyperparams: `Hyperparameters`
        :param agg_info: communication information related to aggregator
        :type agg_info: `dict`
        :param synch: get model update synchronously
        :type synch: `boolean`
        """
        if synch:
            super(PartyProtocolHandlerRabbitMQ, self).__init__(
                fl_model,
                connection,
                data_handler,
                local_training_handler,
                hyperparams,
                agg_info,
                synch,
                is_private,
                **kwargs
            )

        else:
            raise Exception(
                'RabbitMQ connection currently only supports synchronous mode'
            )

    def handle_async_request(self, msg):
        """Handle all incoming requests asynchronously and route it to
        respective methods in local training handler

        :param msg: Message object form connection
        :type msg: `Message`
        :return: Response message sent back to requester
        :rtype: ResponseMessage
        """
        try:
            response_msg = ResponseMessage(
                message_type=MessageType.ACK.value,
                id_request=-1, data={'ACK': True}
            )
            logger.info("received a async request")

            id_request = msg.get_header()['id_request']

            self.execute_async(id_request, msg)

            logger.info("finished async request")

        except Exception as ex:
            logger.info(ex)

        return response_msg
